from __future__ import annotations

import typing

import numpy as np

from prob_conf_mat.stats import summarize_posterior

if typing.TYPE_CHECKING:
    from prob_conf_mat.metrics import Metric, AveragedMetric
    import matplotlib  # noqa: F401
    from matplotlib.figure import Figure

IMPLEMENTED_METHODS = {"kde", "hist", "histogram"}


def distribution_plot(
    study,
    metric: Metric | AveragedMetric,
    class_label: int,
    method: str = "kde",
    bandwidth: float = 1.0,
    bins: int | list[int] | str = "auto",
    normalize: bool = False,
    figsize: typing.Optional[tuple[float, float]] = None,
    fontsize: float = 9,
    axis_fontsize: typing.Optional[float] = None,
    edge_colour: str = "black",
    area_colour: str = "gray",
    area_alpha: float = 0.5,
    plot_median_line: bool = True,
    median_line_colour: str = "black",
    median_line_format: str = "--",
    plot_hdi_lines: bool = True,
    hdi_lines_colour: str = "black",
    hdi_line_format: str = "-",
    plot_obs_point: bool = True,
    obs_point_marker: str = "D",
    obs_point_colour: str = "black",
    obs_point_size: typing.Optional[float] = None,
    plot_extrema_lines: bool = True,
    extrema_lines_colour: str = "black",
    extrema_lines_format: str = "-",
    extrema_line_height: float = 12,
    extrema_lines_width: float = 1,
    plot_base_line: bool = True,
    base_line_colour: str = "black",
    base_line_format: str = "-",
    base_line_width: int = 1,
    plot_experiment_name: bool = True,
) -> Figure:
    """Plots a distribution.

    Args:
        study (Study): the study object calling this function
        metric (Metric | AveragedMetric): the metric
        method (str, optional): the method for displaying a histogram, provided by Seaborn. Can be either a histogram or KDE. Defaults to "kde".
        bandwidth (float, optional): the bandwith parameter for the KDE. Corresponds to [Seaborn's `bw_adjust` parameter](https://seaborn.pydata.org/generated/seaborn.kdeplot.html). Defaults to 1.0.
        bins (int | list[int] | str, optional): the number of bins to use in the histrogram. Corresponds to [numpy's `bins` parameter](https://numpy.org/doc/stable/reference/generated/numpy.histogram_bin_edges.html#numpy.histogram_bin_edges). Defaults to "auto".
        normalize (bool, optional): if normalized, each distribution will be scaled to [0, 1]. Otherwise, uses a shared y-axis. Defaults to False.
        figsize (tuple[float, float], optional): the figure size, in inches. Corresponds to matplotlib's `figsize` parameter. Defaults to None, in which case a decent default value will be approximated.
        fontsize (float, optional): fontsize for the experiment name labels. Defaults to 9.
        axis_fontsize (float, optional): fontsize for the x-axis ticklabels. Defaults to None, in which case the fontsize will be used.
        edge_colour (str, optional): the colour of the histogram or KDE edge. Corresponds to [matplotlib's `color` parameter](https://matplotlib.org/stable/users/explain/colors/colors.html#colors-def). Defaults to "black".
        area_colour (str, optional): the colour of the histogram or KDE filled area. Corresponds to [matplotlib's `color` parameter](https://matplotlib.org/stable/users/explain/colors/colors.html#colors-def). Defaults to "gray".
        area_alpha (float, optional): the opacity of the histogram or KDE filled area. Corresponds to [matplotlib's `alpha` parameter](). Defaults to 0.5.
        plot_median_line (bool, optional): whether to plot the median line. Defaults to True.
        median_line_colour (str, optional): the colour of the median line. Corresponds to [matplotlib's `color` parameter](https://matplotlib.org/stable/users/explain/colors/colors.html#colors-def). Defaults to "black".
        median_line_format (str, optional): the format of the median line. Corresponds to [matplotlib's `linestyle` parameter](https://matplotlib.org/stable/gallery/lines_bars_and_markers/linestyles.html). Defaults to "--".
        plot_hdi_lines (bool, optional): whether to plot the HDI lines. Defaults to True.
        hdi_lines_colour (str, optional): the colour of the HDI lines. Corresponds to [matplotlib's `color` parameter](https://matplotlib.org/stable/users/explain/colors/colors.html#colors-def). Defaults to "black".
        hdi_line_format (str, optional): the format of the HDI lines. Corresponds to [matplotlib's `linestyle` parameter](https://matplotlib.org/stable/gallery/lines_bars_and_markers/linestyles.html). Defaults to "-".
        plot_obs_point (bool, optional): whether to plot the observed value as a marker. Defaults to True.
        obs_point_marker (str, optional): the marker type of the observed value. Corresponds to [matplotlib's `marker` parameter](https://matplotlib.org/stable/gallery/lines_bars_and_markers/marker_reference.html#unfilled-markers). Defaults to "D".
        obs_point_colour (str, optional): the colour of the observed marker. Corresponds to [matplotlib's `color` parameter](https://matplotlib.org/stable/users/explain/colors/colors.html#colors-def). Defaults to "black".
        obs_point_size (float, optional): the size of the observed marker. Defaults to None.
        plot_extrema_lines (bool, optional): whether to plot small lines at the distribution extreme values. Defaults to True.
        extrema_lines_colour (str, optional): the colour of the extrema lines. Defaults to "black".
        extrema_lines_format (str, optional): the format of the extrema lines. Corresponds to [matplotlib's `linestyle` parameter](https://matplotlib.org/stable/gallery/lines_bars_and_markers/linestyles.html). Defaults to "-".
        extrema_line_height (int, optional): the height of the extremea lines in 'points' units. Defaults to 8.
        plot_base_line (bool, optional): whether to plot a line at the base of the distribution. Defaults to True.
        base_line_colour (str, optional): the colour of the base line. Corresponds to [matplotlib's `color` parameter](https://matplotlib.org/stable/users/explain/colors/colors.html#colors-def). Defaults to "black".
        base_line_format (str, optional): the format of the base line. Corresponds to [matplotlib's `linestyle` parameter](https://matplotlib.org/stable/gallery/lines_bars_and_markers/linestyles.html). Defaults to "-".
        plot_experiment_name (bool, optional): whether to plot the experiment names as labels. Defaults to True.

    Returns:
        matplotlib.figure.Figure: the complete figure
    """

    try:
        # Import optional dependencies
        import matplotlib
        import matplotlib.pyplot as plt
        import seaborn as sns

    except ModuleNotFoundError as e:
        raise ModuleNotFoundError(
            f"Visualization requires optional dependencies: [matplotlib, pyplot]. Currently missing: {e}"
        )

    total_num_experiments = sum(map(len, study.experiment_groups.values()))

    if figsize is None:
        # Try to set a decent default figure size
        _figsize = (6.29921, max(0.625 * total_num_experiments, 2.5))
    else:
        _figsize = figsize

    fig, axes = plt.subplots(
        total_num_experiments, 1, figsize=_figsize, sharey=(not normalize)
    )

    if total_num_experiments == 1:
        axes = np.array([axes])

    metric_bounds = metric.bounds

    i = 0

    all_min_x = []
    all_max_x = []
    all_max_height = []
    all_medians = []
    all_hdi_ranges = []
    for experiment_group_name, experiment_group in study.experiment_groups.items():
        for experiment_name, experiment in experiment_group.experiments.items():
            if plot_experiment_name:
                # Set the axis title
                # Needs to happen before KDE
                axes[i].set_ylabel(
                    f"{experiment_group_name}/{experiment_name}",
                    rotation=0,
                    va="center",
                    ha="right",
                    fontsize=fontsize,
                )

            distribution_samples = study.get_metric_samples(
                metric=metric.name,
                experiment_group=experiment_group_name,
                experiment=experiment_name,
                sampling_method="posterior",
            ).values[:, class_label]

            # Get summary statistics
            posterior_summary = summarize_posterior(
                distribution_samples, ci_probability=study.ci_probability
            )

            all_medians.append(posterior_summary.median)
            all_hdi_ranges.append(posterior_summary.hdi[1] - posterior_summary.hdi[0])

            if method not in IMPLEMENTED_METHODS:
                del fig, axes
                raise ValueError(
                    f"Parameter `method` must be one of: {IMPLEMENTED_METHODS}. Currently: {method}"
                )

            elif method == "kde":
                # Plot the kde
                sns.kdeplot(
                    distribution_samples,
                    fill=False,
                    bw_adjust=bandwidth,
                    ax=axes[i],
                    color=edge_colour,
                    clip=metric_bounds,
                    zorder=2,
                )

                kdeline = axes[i].lines[0]

                kde_x = kdeline.get_xdata()
                kde_y = kdeline.get_ydata()

                all_min_x.append(kde_x[0])
                all_max_x.append(kde_x[-1])
                all_max_height.append(np.max(kde_y))

                if area_colour is not None:
                    axes[i].fill_between(
                        kde_x, kde_y, color=area_colour, zorder=0, alpha=area_alpha
                    )

            elif method == "hist" or method == "histogram":
                sns.histplot(
                    distribution_samples,
                    fill=False,
                    bins=bins,
                    stat="density",
                    element="step",
                    ax=axes[i],
                    color=edge_colour,
                    zorder=2,
                )

                kdeline = axes[i].lines[0]

                kde_x = kdeline.get_xdata()
                kde_y = kdeline.get_ydata()

                all_min_x.append(kde_x[0])
                all_max_x.append(kde_x[-1])
                all_max_height.append(np.max(kde_y))

                kde_x = np.repeat(kde_x, 2)
                kde_y = np.concatenate([[0], np.repeat(kde_y, 2)[:-1]])

                if area_colour is not None:
                    axes[i].fill_between(
                        kde_x,
                        kde_y,
                        color=area_colour,
                        zorder=0,
                        alpha=area_alpha,
                    )

            if plot_obs_point:
                # Add a point for the true point value
                observed_metric_value = study.get_metric_samples(
                    metric=metric.name,
                    experiment_group=experiment_group_name,
                    experiment=experiment_name,
                    sampling_method="input",
                ).values[:, class_label]

                axes[i].scatter(
                    observed_metric_value,
                    0,
                    marker=obs_point_marker,
                    color=obs_point_colour,
                    s=obs_point_size,
                    clip_on=False,
                    zorder=2,
                )

            if plot_median_line:
                # Plot median line
                median_x = posterior_summary.median

                y_median = np.interp(
                    x=median_x,
                    xp=kde_x,  # type: ignore
                    fp=kde_y,  # type: ignore
                )  # type: ignore

                axes[i].vlines(
                    median_x,
                    0,
                    y_median,
                    color=median_line_colour,
                    linestyle=median_line_format,
                    zorder=1,
                )

            if plot_hdi_lines:
                x_hdi_lb = posterior_summary.hdi[0]

                y_hdi_lb = np.interp(
                    x=x_hdi_lb,
                    xp=kde_x,  # type: ignore
                    fp=kde_y,  # type: ignore
                )  # type: ignore

                axes[i].vlines(
                    x_hdi_lb,
                    0,
                    y_hdi_lb,
                    color=hdi_lines_colour,
                    linestyle=hdi_line_format,
                    zorder=1,
                )

                x_hdi_ub = posterior_summary.hdi[1]

                y_hdi_ub = np.interp(
                    x=x_hdi_ub,
                    xp=kde_x,  # type: ignore
                    fp=kde_y,  # type: ignore
                )  # type: ignore

                axes[i].vlines(
                    x_hdi_ub,
                    0,
                    y_hdi_ub,
                    color=hdi_lines_colour,
                    linestyle=hdi_line_format,
                    zorder=1,
                )

            i += 1

    smallest_hdi_range_i = int(np.argmin(all_hdi_ranges))

    grand_min_x = max(
        np.min(all_min_x),
        all_medians[smallest_hdi_range_i] - 5 * all_hdi_ranges[smallest_hdi_range_i],
    )
    grand_max_x = min(
        np.max(all_max_x),
        all_medians[smallest_hdi_range_i] + 5 * all_hdi_ranges[smallest_hdi_range_i],
    )

    # Decide on the xlim
    data_range = grand_min_x - grand_max_x
    metric_range = metric_bounds[1] - metric_bounds[0]

    # If the data range spans more than half the metric range
    # Just plot the whole metric range
    if (
        data_range / metric_range > 0.5
        and np.isfinite(metric_bounds[0])
        and np.isfinite(metric_bounds[1])
    ):
        x_lim_min = metric_bounds[0]
        x_lim_max = metric_bounds[1]
    else:
        # If close enough to the metric minimum, use that value
        if (
            np.isfinite(metric_range)
            and (grand_min_x - metric_bounds[0]) / metric_range < 0.05
        ):
            x_lim_min = metric_bounds[0]
        else:
            x_lim_min = grand_min_x  # - 0.05 * (grand_max_x - grand_min_x)

        # If close enough to the metric maximum, use that value
        if (
            np.isfinite(metric_range)
            and (metric_bounds[1] - grand_max_x) / metric_range < 0.05
        ):
            x_lim_max = metric_bounds[1]
        else:
            x_lim_max = grand_max_x  # + 0.05 * (grand_max_x - grand_min_x)

    for ax in axes:
        ax.set_xlim(x_lim_min, x_lim_max)

    for i, ax in enumerate(axes):
        if plot_base_line:
            # Add base line
            ax.hlines(
                0,
                max(all_min_x[i], grand_min_x),
                min(all_max_x[i], grand_max_x),
                color=base_line_colour,
                ls=base_line_format,
                linewidth=base_line_width,
                zorder=3,
                clip_on=False,
            )

        standard_length = (
            ax.transData.inverted().transform([0, extrema_line_height])[1]
            - ax.transData.inverted().transform([0, 0])[1]
        )

        if plot_extrema_lines:
            # Add lines for the horizontal extrema
            if all_min_x[i] >= grand_min_x:
                ax.vlines(
                    all_min_x[i],
                    0,
                    standard_length,
                    color=extrema_lines_colour,
                    ls=extrema_lines_format,
                    linewidth=extrema_lines_width,
                    zorder=3,
                    clip_on=False,
                )

            if all_max_x[i] <= grand_max_x:
                ax.vlines(
                    all_max_x[i],
                    0,
                    standard_length,
                    color=extrema_lines_colour,
                    ls=extrema_lines_format,
                    zorder=3,
                    linewidth=extrema_lines_width,
                    clip_on=False,
                )

        # Remove the ticks
        ax.set_xticks([])
        ax.set_yticks([])

        # Remove the axis spine
        ax.spines["top"].set_visible(False)
        ax.spines["right"].set_visible(False)
        ax.spines["left"].set_visible(False)
        ax.spines["bottom"].set_visible(False)

    # Add the axes back, but only for the bottom plot
    axes[-1].spines["bottom"].set_visible(True)
    axes[-1].xaxis.set_major_locator(matplotlib.ticker.AutoLocator())  # type: ignore
    axes[-1].set_yticks([])
    axes[-1].tick_params(
        axis="x", labelsize=axis_fontsize if axis_fontsize is not None else fontsize
    )

    fig.tight_layout()

    return fig



from __future__ import annotations

import typing
import collections.abc

import numpy as np

class RNG:

    def __init__(self, seed: int | typing.Sequence[int] | None, position: tuple = ()) -> None:
        """A wrapper for Numpy's Generator class, that tracks any spawned child RNGs.

        This way, we can change the RNG in a reproducible manner, on the fly.

        Args:
            seed (int | typing.Sequence[int] | None): A seed to initialize the `BitGenerator`. If None, then fresh, unpredictable entropy will be pulled from the OS. If an `int` or `array_like[ints]` is passed, then all values must be non-negative and will be passed to `SeedSequence` to derive the initial `BitGenerator` state. One may also pass in a `SeedSequence` instance. Additionally, when passed a `BitGenerator`, it will be wrapped by `Generator`. If passed a `Generator`, it will be returned unaltered. When passed a legacy `RandomState` instance it will be coerced to a `Generator`.
            position (tuple, optional): the position of the RNG in the tree. Typically this is not set by the user. Defaults to a root node.
        """

        # The current seed & position
        self._seed = seed
        self.position = position

        # The current seed sequence
        self.seed_sequence: np.random.SeedSequence = np.random.SeedSequence(entropy=self._seed, spawn_key=self.position)

        # The RNG object
        self.rng: np.random.Generator = np.random.default_rng(seed=self.seed_sequence)

        # Any RNGs spawned from this RNG
        self.children: list[typing.Self] = []

    def __getattr__(self, name):
        """Called for failed attribute accesses."""
        return self.rng.__getattribute__(name)

    def spawn(self, n_children: int) -> list[typing.Self]: # type: ignore
        """Spawns a new independent RNG. It has a position lower in the tree, but shares the seed of its parent.

        Args:
            n_children (int): the number of children to spawn

        Returns:
            list[typing.Self]: a list of the newly generated child RNGs
        """

        cur_num_children = len(self.children)

        spawned_rngs: list[typing.Self] = []
        for i in range(n_children):
            new_rng = RNG(seed=self.seed, position=(*self.position, cur_num_children+i))

            spawned_rngs.append(new_rng) # type: ignore

        self.children.extend(spawned_rngs)

        return spawned_rngs

    @property
    def seed(self) -> int | typing.Sequence[int] | None:
        return self._seed

    @seed.setter
    def seed(self, value: int | typing.Sequence[int] | None) -> None:
        """Changes the seed without changing the position.

        Args:
            value (int | Sequence[int] | None): the need seed value
        """

        self._seed = value

        self.seed_sequence = np.random.SeedSequence(entropy=self._seed, spawn_key=self.position)

        self.rng = np.random.default_rng(seed=self.seed_sequence)

        for child in self.children:
            child.seed = value

    def __repr__(self) -> str:
        return f"RNG(seed={self.seed}, position={self.position})"

    def __str__(self) -> str:
        return f"RNG(seed={self.seed}, position={self.position})"



from collections import OrderedDict
import hashlib
import pickle
import time
import typing
import warnings

import numpy as np

from prob_conf_mat.metrics import get_metric
from prob_conf_mat.experiment_aggregation import get_experiment_aggregator
from prob_conf_mat.io import validate_confusion_matrix
from prob_conf_mat.stats import _DIRICHLET_PRIOR_STRATEGIES
from prob_conf_mat.utils import RNG


# TODO: document this class
class ConfigWarning(Warning):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)


# TODO: document this class
class ConfigError(Exception):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)


# TODO: document this class
class Config:
    def __init__(
        self,
        seed: int | None = None,
        num_samples: int | None = None,
        ci_probability: float | None = None,
        experiments: dict[str, dict[str, dict[str, typing.Any]]] = {},
        metrics: dict[str, dict[str, typing.Any]] = {},
    ) -> None:
        # Set the RNG
        # Allows for potentially updating the seed
        self.rng = RNG(seed=None)

        # Set the initial values
        self.__setattr__("seed", seed)
        self.__setattr__("num_samples", num_samples)
        self.__setattr__("ci_probability", ci_probability)
        self.__setattr__("experiments", experiments)
        self.__setattr__("metrics", metrics)

    @property
    def seed(self) -> int:
        return self._seed

    def _validate_seed(self, value: int | None) -> int:
        # Handle default seed
        if value is None:
            value = int(time.time() * 256)

            warnings.warn(
                f"Recieved `None` as seed. Defaulting to fractional seconds: {value}",
                category=ConfigWarning,
            )

        # Handle seed of wrong type
        else:
            if not isinstance(value, int):
                try:
                    initial_value_type = type(value)
                    value = int(value)

                    warnings.warn(
                        (
                            f"Parameter `seed` must be a positive integer. "
                            f"Received: {initial_value_type}. Parsed as: {value}"
                        ),
                        category=ConfigWarning,
                    )

                except Exception as e:  # noqa: BLE001
                    raise TypeError(
                        (
                            f"Parameter `seed` must be a positive integer. "
                            f"Currently: {type(value)}. While trying to convert"
                            f"encountered the following exception: {e}"
                        ),
                    )

        # Validate seed value
        if value < 0:
            raise ConfigError(
                f"Parameter `seed` must be a positive integer. Currently: {value}.",
            )

        return value

    @seed.setter
    def seed(self, value: int | None) -> None:
        value_: int = self._validate_seed(value=value)

        self._seed = value_
        self.rng.seed = self._seed

    @property
    def num_samples(self) -> int:
        return self._num_samples

    def _validate_num_samples(self, value: int | None) -> int:
        # Handle default parameter
        if value is None:
            warnings.warn(
                message=(
                    "Parameter `num_samples` is `None`. "
                    "Setting to default value of 10000. "
                    "This value is arbitrary, however, and should be carefully considered."
                ),
                category=ConfigWarning,
            )

            value = 10000

        # Handle parameter of wrong type
        else:
            if not isinstance(value, int):
                try:
                    initial_value_type = type(value)
                    value = int(value)

                    warnings.warn(
                        (
                            f"Parameter `num_samples` must be a strictly positive integer. "
                            f"Received: {initial_value_type}. Parsed as: {value}"
                        ),
                        category=ConfigWarning,
                    )

                except Exception as e:  # noqa: BLE001
                    raise TypeError(
                        (
                            f"Parameter `num_samples` must be a strictly positive integer. "
                            f"Currently: {type(value)}. While trying to convert encountered "
                            f"the following exception: {e}"
                        ),
                    )

        # Validate num_samples value
        if value <= 0:
            raise ConfigError(
                f"Parameter `num_samples` must be greater than 0. Currently: {value}",
            )

        # TODO: consider increasing the recommended value to from 10e+4 to 10e+5
        if value < 10000:
            warnings.warn(
                message=(
                    f"Parameter `num_samples` should be large to reduce variability. "
                    f"Consider increasing. Currently: {value}"
                ),
                category=ConfigWarning,
            )

        return value

    @num_samples.setter
    def num_samples(self, value: int) -> None:
        value_ = self._validate_num_samples(value=value)

        self._num_samples = value_

    @property
    def ci_probability(self) -> float:
        return self._ci_probability

    def _validate_ci_probability(self, value: float | None) -> float:
        # Handle default parameter
        if value is None:
            warnings.warn(
                message=(
                    "Parameter `ci_probability` is `None`. Setting to default value of 0.95. "
                    "This value is arbitrary, however, and should be carefully considered."
                ),
                category=ConfigWarning,
            )

            value = 0.95

        # Handle parameter of wrong type
        else:
            if not isinstance(value, float):
                try:
                    initial_value_type = type(value)
                    value = float(value)

                    warnings.warn(
                        (
                            f"Parameter `ci_probability` must be a float. "
                            f"Received: {initial_value_type}. Parsed as: {value}"
                        ),
                        category=ConfigWarning,
                    )

                except Exception as e:  # noqa: BLE001
                    raise TypeError(
                        (
                            "Parameter `ci_probability` must be a float. "
                            f"Currently: {type(value)}. "
                            f"While trying to convert encountered the following exception: {e}"
                        ),
                    )

        # Validate ci_probability value
        if not (value > 0.0 and value <= 1.0):
            raise ConfigError(
                f"Parameter `ci_probability` must be within (0.0, 1.0]. Currently: {value}",
            )

        return value

    @ci_probability.setter
    def ci_probability(self, value: float) -> None:
        value_ = self._validate_ci_probability(value=value)

        self._ci_probability = value_

    @property
    def experiments(self) -> dict[str, dict[str, dict[str, typing.Any]]]:
        return self._experiments

    def _validate_experiments(
        self,
        value: dict[str, dict[str, dict[str, typing.Any]]] | None,
    ) -> dict[str, dict[str, dict[str, typing.Any]]]:
        def _validate_single_experiment(
            experiment_name: str,
            experiment_config: dict[str, typing.Any],
        ) -> dict[str, typing.Any]:
            updated_experiment_config = dict()

            # ==========================================================================
            # Confusion matrix
            # Expected type: Int[ArrayLike,'num_classes num_classes']
            # ==========================================================================
            if "confusion_matrix" not in experiment_config:
                raise ConfigError(
                    f"Experiment '{experiment_name}' has no confusion matrix. Please add one.",
                )
            confusion_matrix = experiment_config["confusion_matrix"]

            confusion_matrix = validate_confusion_matrix(
                confusion_matrix=confusion_matrix,
            )

            updated_experiment_config["confusion_matrix"] = confusion_matrix

            # ==========================================================================
            # Prevalence prior
            # Expected type: str | float | Float[ArrayLike,'num_classes']
            # ==========================================================================
            # First check if the key is in there, and fall back to standard default if not
            prevalence_prior = experiment_config.get("prevalence_prior", None)

            # Handle default parameter
            if prevalence_prior is None:
                warnings.warn(
                    (
                        f"Experiment '{experiment_name}'s prevalence prior is `None`. "
                        f"Defaulting to the 0 (Haldane) prior."
                    ),
                    category=ConfigWarning,
                )

                prevalence_prior = 0.0

            # Check if the string is a valid strategy
            elif isinstance(prevalence_prior, str):
                if prevalence_prior not in _DIRICHLET_PRIOR_STRATEGIES:
                    raise ConfigError(
                        (
                            f"Experiment '{experiment_name}'s prevalence prior is invalid. "
                            f"Currently: {prevalence_prior}. If `str`, must be one of: "
                            f"{set(_DIRICHLET_PRIOR_STRATEGIES.keys())}"
                        ),
                    )

            # Accept positive integer and float values
            elif isinstance(prevalence_prior, int | float):
                if prevalence_prior < 0:
                    raise ConfigError(
                        (
                            f"Experiment '{experiment_name}'s prevalence prior is invalid. "
                            f"Currently: {prevalence_prior}. If numeric, must be greater than 0."
                        ),
                    )

            # Try to convert anything else to a numpy array
            else:
                try:
                    prevalence_prior = np.array(prevalence_prior, dtype=np.float64)
                except Exception as e:  # noqa: BLE001
                    raise ConfigError(
                        (
                            f"Experiment '{experiment_name}'s prevalence prior is invalid. "
                            f"Expecting a type of str | float | Float[ArrayLike,'num_classes']. "
                            f"Currently: {type(prevalence_prior)}. "
                            f"While trying to convert to `np.ndarray`, the following exception "
                            f"was encountered: {e}"
                        ),
                    )

                # Additional numpy array validation
                # Check shape against confusion matrix
                if (
                    len(prevalence_prior.shape) != 1
                    or prevalence_prior.shape[0] != confusion_matrix.shape[0]
                ):
                    try:
                        prevalence_prior = prevalence_prior.reshape(
                            (confusion_matrix.shape[0],),
                        )
                    except Exception as e:  # noqa: BLE001
                        raise ConfigError(
                            (
                                f"Experiment '{experiment_name}'s prevalence prior is malformed. "
                                f"Expecting a 1D vector with length equal to the number of classes "
                                f"in the experiment. Current shape: {prevalence_prior.shape}. "
                                f"Expecting shape: {(confusion_matrix.shape[0],)}. "
                                f"While trying to reshape, encountered the following exception: {e}"
                            ),
                        )

                # Check that values are all positive
                if not np.all(prevalence_prior > 0.0):
                    raise ConfigError(
                        (
                            f"Experiment '{experiment_name}'s prevalence prior is invalid. "
                            f"If providing an arraylike of values, all values must be positive."
                        ),
                    )

            updated_experiment_config["prevalence_prior"] = prevalence_prior

            # ==========================================================================
            # Confusion prior
            # Expected type: str | float | Float[ArrayLike,'num_classes num_classes']
            # ==========================================================================
            # First check if the key is in there, and fall back to standard default if not
            confusion_prior = experiment_config.get("confusion_prior", None)

            # Handle default parameter value
            if confusion_prior is None:
                warnings.warn(
                    (
                        f"Experiment '{experiment_name}'s confusion prior is `None`. "
                        f"Defaulting to the 0 (Haldane) prior."
                    ),
                    category=ConfigWarning,
                )

                confusion_prior = 0.0

            elif isinstance(confusion_prior, str):
                if confusion_prior not in _DIRICHLET_PRIOR_STRATEGIES:
                    raise ConfigError(
                        (
                            f"Experiment '{experiment_name}'s confusion prior is invalid. "
                            f"Currently: {confusion_prior}. If `str`, must be one of: "
                            f"{set(_DIRICHLET_PRIOR_STRATEGIES.keys())}",
                        ),
                    )

            # Accept positive integer and float values
            elif isinstance(confusion_prior, int | float):
                if confusion_prior < 0:
                    raise ConfigError(
                        (
                            f"Experiment '{experiment_name}'s confusion prior is invalid. "
                            f"Currently: {confusion_prior}. If numeric, must be greater than 0."
                        ),
                    )

            else:
                try:
                    confusion_prior = np.array(confusion_prior, dtype=np.float64)
                except Exception as e:  # noqa: BLE001
                    raise ConfigError(
                        (
                            f"Experiment '{experiment_name}'s confusion prior is invalid. "
                            f"Expecting a type of str | float | Float[ArrayLike,'num_classes']. "
                            f"Currently: {type(confusion_prior)}. While trying to convert to "
                            f"`np.ndarray`, the following exception was encountered: {e}"
                        ),
                    )

                # Check shape against confusion matrix
                if (
                    (len(confusion_prior.shape) != 2)
                    or (confusion_prior.shape[0] != confusion_matrix.shape[0])
                    or (confusion_prior.shape[1] != confusion_matrix.shape[1])
                    or (confusion_prior.shape[0] != confusion_prior.shape[1])
                ):
                    try:
                        confusion_prior = confusion_prior.reshape(
                            confusion_matrix.shape,
                        )
                    except Exception as e:  # noqa: BLE001
                        raise ConfigError(
                            (
                                f"Experiment '{experiment_name}'s confusion prior is malformed. "
                                f"Expecting a square 2D matrix with length equal to the number of "
                                f"classes in the experiment. "
                                f"Current shape: {confusion_prior.shape}. "
                                f"Expecting shape: {confusion_matrix.shape}. "
                                f"While trying to reshape, encountered the following exception: {e}"
                            ),
                        )

                # Check that values are all positive
                if not np.all(confusion_prior > 0.0):
                    raise ConfigError(
                        (
                            f"Experiment '{experiment_name}'s confusion prior is invalid. "
                            f"If providing an arraylike of values, all values must be positive."
                        ),
                    )

            updated_experiment_config["confusion_prior"] = confusion_prior

            # ==========================================================================
            # Kwargs
            # Not expecting any of these for now
            # ==========================================================================
            kwargs = {
                k: v
                for k, v in experiment_config.items()
                if k not in ["confusion_matrix", "prevalence_prior", "confusion_prior"]
            }

            if len(kwargs) != 0:
                warnings.warn(
                    (
                        f"Experiment '{experiment_name}'s received the following "
                        f"superfluous parameters: {set(kwargs.keys())}. "
                        f"These are currently just ignored."
                    ),
                    category=ConfigWarning,
                )

            updated_experiment_config.update(kwargs)

            return updated_experiment_config

        # Handle default value
        experiments_config = dict() if value is None else value

        # Duck type to make sure it matches dict protocol
        if not (
            hasattr(experiments_config, "get") and hasattr(experiments_config, "items")
        ):
            raise ConfigError(
                (
                    f"The experiments configuration must implement the `get` and `items` "
                    f"attributes like a `dict`. Current type: {type(value)}"
                ),
            )

        updated_experiments_config = dict()
        for experiment_group_name, experiment_group in experiments_config.items():
            # ==================================================================
            # Experiment group name
            # ==================================================================
            if experiment_group_name == "aggregated":
                raise ConfigError("An experiment group may not be named 'aggregated'.")

            if not isinstance(experiment_group_name, str):
                try:
                    experiment_group_name = str(experiment_group_name)
                except Exception as e:  # noqa: BLE001
                    raise ConfigError(
                        (
                            f"Experiment group `{experiment_group_name}` must be an instance of "
                            f"`str`, but got `{type(experiment_group_name)}`. "
                            f"While trying to convert, ran into the following exception: {e}"
                        ),
                    )

            # Duck type to make sure it matches dict protocol
            if not (
                hasattr(experiment_group, "get") and hasattr(experiment_group, "items")
            ):
                raise ConfigError(
                    (
                        f"The experiment group configuration must implement the `get` and `items` "
                        f"attributes like a `dict`. Currently: {type(experiment_group)}"
                    ),
                )

            # ==================================================================
            # Experiment group
            # ==================================================================
            confusion_matrix_shapes = set()
            updated_experiment_group_config = dict()
            for experiment_name, experiment_config in experiment_group.items():
                # ==============================================================
                # Experiment name
                # ==============================================================
                if experiment_name == "aggregated":
                    raise ConfigError("An experiment may not be named 'aggregated'.")

                if not isinstance(experiment_name, str):
                    try:
                        experiment_name = str(experiment_name)
                    except Exception as e:  # noqa: BLE001
                        raise ConfigError(
                            (
                                f"The key for `{experiment_group_name}/{experiment_name}` "
                                f"must be an instance of `str`, but got "
                                f"`{type(experiment_group_name)}`. "
                                f"While trying to convert, ran into the following exception: {e}"
                            ),
                        )

                # ==============================================================
                # Experiment
                # ==============================================================
                full_experiment_name = f"{experiment_group_name}/{experiment_name}"
                updated_experiment_config = _validate_single_experiment(
                    experiment_name=full_experiment_name,
                    experiment_config=experiment_config,
                )

                updated_experiment_group_config[experiment_name] = (
                    updated_experiment_config
                )

                confusion_matrix_shapes.add(
                    updated_experiment_config["confusion_matrix"].shape,
                )

            if len(confusion_matrix_shapes) != 1:
                raise ConfigError(
                    (
                        f"Experiment group '{experiment_group_name}' has "
                        f"incongruent confusion matrices. "
                        f"Found shapes: {confusion_matrix_shapes}"
                    ),
                )

            updated_experiments_config[experiment_group_name] = (
                updated_experiment_group_config
            )

        return updated_experiments_config

    @experiments.setter
    def experiments(
        self,
        value: dict[str, dict[str, dict[str, typing.Any]]],
    ) -> None:
        value = self._validate_experiments(value)

        self._experiments = value

    @property
    def num_experiments(self) -> int:
        return sum(map(len, self.experiments.values()))

    @property
    def num_experiment_groups(self) -> int:
        return len(self.experiments)

    @property
    def metrics(self) -> dict[str, dict[str, typing.Any]]:
        return self._metrics

    def _validate_metrics(
        self,
        value: dict[str, dict[str, typing.Any]],
    ) -> dict[str, dict[str, typing.Any]]:
        def validate_metric_configuration(key: str, configuration: dict) -> None:
            # Empty configuration is allowed
            if len(configuration) == 0:
                return

            #! If non-empty, must contain an aggregation key
            if "aggregation" not in configuration:
                raise ConfigError(
                    (
                        f"The metric configuration for {key} must contain an `aggregation` key. "
                        f"Currently: {configuration}"
                    ),
                )

            #! Aggregation key must map to registered aggregation
            #! Aggregation config must be valid
            try:
                kwargs = {k: v for k, v in configuration.items() if k != "aggregation"}
                get_experiment_aggregator(
                    configuration["aggregation"],
                    rng=RNG(None),
                    **kwargs,
                )

            except Exception as e:  # noqa: BLE001
                raise ConfigError(
                    (
                        f"The aggregation configuration for metric {key} is invalid. "
                        f"Currently: {configuration}. "
                        f"While trying to parse, the following exception was encountered: {e}"
                    ),
                )

        # If not yet initialized, initialize to an empty dictionary
        if value is None:
            return dict()

        if not (hasattr(value, "get") and hasattr(value, "items")):
            raise ConfigError(
                (
                    f"Metrics configuration must implement the `get` and `items` "
                    f"attributes like a `dict`. "
                    f"Current type: {type(value)}"
                ),
            )

        default_config = value.get("__default__", dict())

        # Validate the default config
        if len(default_config) != 0:
            validate_metric_configuration("__default__", default_config)

        updated_metrics_config = OrderedDict()
        for metric_key, metric_config in value.items():
            # Do not validate the __default__ config
            if metric_key == "__default__":
                continue

            # Validate type ====================================================
            if not isinstance(metric_key, str):
                try:
                    metric_key = str(metric_key)
                except Exception as e:  # noqa: BLE001
                    raise ConfigError(
                        (
                            f"The keys in metrics must of type `str`. "
                            f"Currently: {type(metric_key)}. "
                            f"While trying to convert, the following exception was encountered: {e}"
                        ),
                    )

            if not (hasattr(value, "get") and hasattr(value, "items")):
                raise ConfigError(
                    (
                        f"Configuration for metric {metric_key} must implement the "
                        f"`get` and `items` attributes like a `dict`. "
                        f"Current type: {type(value)}"
                    ),
                )

            # Validate the key of the metric config ============================
            try:
                get_metric(metric_key)
            except Exception as e:  # noqa: BLE001
                raise ConfigError(
                    (
                        f"The following metric is an invalid metric syntax string: `{metric_key}`. "
                        f"While trying to parse, the following exception was encountered: {e}"
                    ),
                )

            # Validate the metric config =======================================
            if len(metric_config) == 0 and len(default_config) == 0:
                # If no metric aggregation config has been passed, make sure
                # there are not more than 1 experiment groups
                if (
                    len(self.experiments) > 0
                    and max(map(len, self.experiments.values())) > 1
                ):
                    # Check for when requesting to aggregate
                    # Allow for studies where the user does not want to aggregate
                    warnings.warn(
                        message=(
                            f"There is an experiment group with multiple experiments, "
                            f"but no aggregation method is provided for metric `{metric_key}`."
                        ),
                        category=ConfigWarning,
                    )

                updated_metrics_config[metric_key] = dict()

            elif len(metric_config) == 0 and len(default_config) != 0:
                # No metric aggregation config has been passed, but a default
                # metric aggregation dict does exist
                updated_metrics_config[metric_key] = default_config

            else:
                # Otherwise, validate each metric aggregation configuration
                validate_metric_configuration(
                    key=metric_key,
                    configuration=metric_config,
                )

                updated_metrics_config[metric_key] = metric_config

        return updated_metrics_config

    @metrics.setter
    def metrics(self, value: dict[str, dict[str, typing.Any]]) -> None:
        value = self._validate_metrics(value=value)

        self._metrics = value

    def to_dict(self) -> dict[str, typing.Any]:
        """Returns the Config as a Pythonic dict."""
        # Necessary to make sure all types are Pythonic
        updated_experiments_config = dict()
        for experiment_group_name, experiment_group in self.experiments.items():
            updated_experiment_group_config = dict()
            for experiment_name, experiment_config in experiment_group.items():
                updated_experiment_config = dict()

                updated_experiment_config["confusion_matrix"] = experiment_config[
                    "confusion_matrix"
                ].tolist()

                if isinstance(experiment_config["prevalence_prior"], np.ndarray):
                    updated_experiment_config["prevalence_prior"] = experiment_config[
                        "prevalence_prior"
                    ].tolist()
                else:
                    updated_experiment_config["prevalence_prior"] = experiment_config[
                        "prevalence_prior"
                    ]

                if isinstance(experiment_config["confusion_prior"], np.ndarray):
                    updated_experiment_config["confusion_prior"] = experiment_config[
                        "confusion_prior"
                    ].tolist()
                else:
                    updated_experiment_config["confusion_prior"] = experiment_config[
                        "confusion_prior"
                    ]

                updated_experiment_group_config[experiment_name] = (
                    updated_experiment_config
                )

            updated_experiments_config[experiment_group_name] = (
                updated_experiment_group_config
            )

        # Final state dict
        state_dict = {
            "seed": int(self.seed),
            "num_samples": int(self.num_samples),
            "ci_probability": float(self.ci_probability),
            "experiments": updated_experiments_config,
            "metrics": dict(self.metrics),
        }

        return state_dict

    @classmethod
    def from_dict(cls, config_dict: dict[str, typing.Any]) -> typing.Self:
        raise NotImplementedError

    @property
    def fingerprint(self) -> str:
        hasher = hashlib.sha256()
        hasher.update(pickle.dumps(self.to_dict()))

        fingerprint = hasher.hexdigest()

        return fingerprint

    def __hash__(self) -> int:
        return self.fingerprint.__hash__()



import typing
from collections import defaultdict
from dataclasses import dataclass
from pathlib import Path


@dataclass
class NotInCache:
    pass


class NestedCache:
    """Recursive defaultdict for storing stuff in a nested dict.

    Examples:
        >>> cache = NestedCache()

        >>> cache.load(['foo', 'bar'])
        KeyError: "Keys ['foo', 'bar'] not in cache"

        >>> cache.cache(['foo', 'bar'], 'baz')

        >>> cache.load(['foo', 'bar'])
        'baz'

    """

    def __init__(self):
        self._cache = self.nested_dict()
        self.fingerprint = None

    def nested_dict(self):
        return defaultdict(self.nested_dict)

    def __getitem__(self, key):
        return self._cache[key]

    def _get(self, dd, keys, i: int):
        # Recursively look up elements in cache
        # First check current layer
        obj = dd.get(keys[i], NotInCache)

        # If key not in cache, or we've exhausted the
        # list of keys, return what came out of the cache
        if obj is NotInCache or i == len(keys) - 1:
            return obj

        # Otherwise, keep going down the cache until we
        # exhaust the list of keys
        else:
            return self._get(obj, keys=keys, i=i + 1)

    def load(self, fingerprint: str, keys, default=NotInCache):
        if fingerprint != self.fingerprint:
            return default

        # Safe method for getting value without altering
        # the cache
        result = self._get(dd=self._cache, keys=keys, i=0)

        if result is NotInCache:
            return default
        else:
            return result

    def _set(self, dd, value, keys, i: int):
        # Recursively set elements in cache
        # If we've exhausted the list of keys,
        # set the value
        if i == len(keys) - 1:
            dd[keys[i]] = value

        # Otherwise, keep going down the cache until we
        # exhaust the list of keys
        else:
            self._set(dd[keys[i]], value=value, keys=keys, i=i + 1)

    def cache(self, fingerprint: str, keys: typing.List[typing.Any], value):
        if fingerprint != self.fingerprint:
            self.clean()
            self.fingerprint = fingerprint

        # Method for setting value
        self._set(dd=self._cache, keys=keys, value=value, i=0)

    def clean(self):
        raise NotImplementedError

    def clear(self):
        self.clean()

    def empty(self):
        self.clean()

    def isin(self, fingerprint: str, keys: typing.List[typing.Any]):
        if fingerprint != self.fingerprint:
            return False

        result = self._get(dd=self._cache, keys=keys, i=0)

        if result is NotInCache:
            return False
        else:
            return True

class InMemoryCache(NestedCache):
    def clean(self):
        self._cache = self.nested_dict()

# TODO: add support for filesystem caching
# class PickleCache(NestedCache):
#     def __init__(self):
#         raise NotImplementedError

#     def clean(self):
#         def _clean_directory(dir: Path):
#             for obj in dir.glob("*"):
#                 if obj.is_dir():
#                     _clean_directory(obj)
#                 else:
#                     obj.unlink()

#             dir.rmdir()

#         if isinstance(self.cache_location, Path) and self.cache_location.exists():
#             # warnings.warn(f"Cleaning cache dir: {self.cache_location}")

#             _clean_directory(self.cache_location)
#         else:
#             raise ValueError("Could not clean cache dir, does not exist.")



from __future__ import annotations
import typing

if typing.TYPE_CHECKING:
    import jaxtyping as jtyping

import numpy as np


def numpy_batched_arithmetic_mean(
    array: jtyping.Float[np.ndarray, "... axis ..."],
    axis: int = -1,
    *,
    keepdims: bool = True,
) -> jtyping.Float[np.ndarray, "... 1 ..."]:
    """Computes the [arithmetic mean](https://en.wikipedia.org/wiki/Arithmetic_mean) over an axis."""  # noqa: E501
    return np.mean(array, axis=axis, keepdims=keepdims)


def numpy_batched_convex_combination(
    array: jtyping.Float[np.ndarray, "... axis ..."],
    axis: int = -1,
    *,
    convex_weights: jtyping.Float[np.ndarray, " num_samples final_dim"],
    keepdims: bool = True,
) -> jtyping.Float[np.ndarray, "... 1 ..."]:
    """Computes a [weighted mean](https://en.wikipedia.org/wiki/Weighted_arithmetic_mean) over an axis."""  # noqa: E501
    return np.sum(convex_weights * array, axis=axis, keepdims=keepdims)


def numpy_batched_harmonic_mean(
    array: jtyping.Float[np.ndarray, "... axis ..."],
    axis: int = -1,
    *,
    keepdims: bool = True,
) -> jtyping.Float[np.ndarray, "... 1 ..."]:
    """Computes the [harmonic mean](https://en.wikipedia.org/wiki/Harmonic_mean) over an axis."""  # noqa: E501
    return np.power(np.mean(np.power(array, -1), axis=axis, keepdims=keepdims), -1)


def numpy_batched_geometric_mean(
    array: jtyping.Float[np.ndarray, "... axis ..."],
    axis: int = -1,
    *,
    keepdims: bool = True,
) -> jtyping.Float[np.ndarray, "... 1 ..."]:
    """Computes the [weighted mean](https://en.wikipedia.org/wiki/Geometric_mean) over an axis."""  # noqa: E501
    return np.exp(np.mean(np.log(array), axis=axis, keepdims=keepdims))



import pytest

from prob_conf_mat.study import Study
from prob_conf_mat.experiment_group import ExperimentGroup
from prob_conf_mat.experiment import Experiment


class TestStudy:
    def test_getitem(self):
        # First define a common setup
        study = Study(seed=0, num_samples=10000, ci_probability=0.95)

        study.add_experiment(
            "test/test_a",
            confusion_matrix=[[1, 0], [0, 1]],
            prevalence_prior=0,
            confusion_prior=0,
        )
        study.add_experiment(
            "test/foo",
            confusion_matrix=[[1, 0], [0, 1]],
            prevalence_prior=0,
            confusion_prior=0,
        )

        # Should fetch at the right level
        assert isinstance(study["test"], ExperimentGroup)
        assert isinstance(study["test/test_a"], Experiment)
        assert isinstance(study["test/foo"], Experiment)

        # Should raise an error when trying to fetch non-existent experiment group
        with pytest.raises(
            KeyError,
            match=r"No experiment group with name .* is currently present",
        ):
            study["foo"]

        with pytest.raises(
            KeyError,
            match=r"No experiment with name .* is currently present",
        ):
            study["test/bar"]

        # Should only accept strings
        with pytest.raises(
            TypeError,
        ):
            study[0]  # type: ignore

        # Should only accept strings
        with pytest.raises(
            ValueError,
        ):
            study["foo/bar/baz"]

        # Test unknown experiment group
        with pytest.raises(
            KeyError,
        ):
            study["foo/bar"]

    def test_init(self):
        # First experiment, then metric
        study = Study(seed=0, num_samples=10000, ci_probability=0.95)

        study.add_experiment(
            "test/test_a",
            confusion_matrix=[[1, 0], [0, 1]],
            prevalence_prior=0,
            confusion_prior=0,
        )

        study.add_metric(metric="acc", aggregation="fe_gaussian")

        # First metric, then experiment
        study = Study(seed=0, num_samples=10000, ci_probability=0.95)

        study.add_metric(metric="acc", aggregation="fe_gaussian")

        study.add_experiment(
            "test/test_a",
            confusion_matrix=[[1, 0], [0, 1]],
            prevalence_prior=0,
            confusion_prior=0,
        )

        # Experiment in __init__
        study = Study(
            seed=0,
            num_samples=10000,
            ci_probability=0.95,
            experiments={
                "test": {
                    "test": {
                        "confusion_matrix": [[1, 0], [0, 1]],
                        "prevalence_prior": 0,
                        "confusion_prior": 0,
                    },
                },
            },
        )

        study.add_metric(metric="acc", aggregation="fe_gaussian")

        # Metric in __init__
        study = Study(
            seed=0,
            num_samples=10000,
            ci_probability=0.95,
            metrics={
                "acc": {"aggregation": "fe_gaussian"},
            },
        )

        study.add_experiment(
            "test/test_a",
            confusion_matrix=[[1, 0], [0, 1]],
            prevalence_prior=0,
            confusion_prior=0,
        )

        # Experiment and metric in __init__
        study = Study(
            seed=0,
            num_samples=10000,
            ci_probability=0.95,
            experiments={
                "test": {
                    "test": {
                        "confusion_matrix": [[1, 0], [0, 1]],
                        "prevalence_prior": 0,
                        "confusion_prior": 0,
                    },
                },
            },
            metrics={
                "acc": {"aggregation": "fe_gaussian"},
            },
        )

    def test_metric_label_validation(self):
        study = Study(seed=0, num_samples=10000, ci_probability=0.95)

        study.add_experiment(
            "test/test_a",
            confusion_matrix=[[1, 0], [0, 1]],
            prevalence_prior=0,
            confusion_prior=0,
        )

        study.add_metric(metric="acc", aggregation="normal")
        study.add_metric(metric="f1", aggregation="normal")

        # Test valid multiclass label
        study._validate_metric_class_label_combination(
            metric="acc",
            class_label=0,
        )

        study._validate_metric_class_label_combination(
            metric="acc",
            class_label=None,
        )

        # Test accessing invalid class
        with pytest.warns(match="Metric is multiclass, ignoring class label."):
            study._validate_metric_class_label_combination(
                metric="acc",
                class_label=1,
            )

        # Test valid binary label
        study._validate_metric_class_label_combination(
            metric="f1",
            class_label=0,
        )

        study._validate_metric_class_label_combination(
            metric="f1",
            class_label=1,
        )

        # Test accessing invalid class
        with pytest.raises(
            ValueError,
            match="is not multiclass. You must provide a class label.",
        ):
            study._validate_metric_class_label_combination(
                metric="f1",
                class_label=None,
            )

        with pytest.raises(ValueError, match="Class label must be in range"):
            study._validate_metric_class_label_combination(
                metric="f1",
                class_label=-1,
            )

        with pytest.raises(ValueError, match="Class label must be in range"):
            study._validate_metric_class_label_combination(
                metric="f1",
                class_label=3,
            )

        # Test unknown metric
        with pytest.raises(KeyError, match="Could not find metric"):
            study._validate_metric_class_label_combination(
                metric="foobarbaz",
                class_label=None,
            )



from __future__ import annotations
import typing

if typing.TYPE_CHECKING:
    import scipy
    import numpy as np
    import jaxtyping as jtyping

    from prob_conf_mat.utils.rng import RNG


def truncated_sample(
    sampling_distribution: scipy.stats.rv_continuous,
    bounds: tuple[float, float],
    rng: RNG,
    num_samples: int,
) -> jtyping.Float[np.ndarray, " num_samples"]:
    """Generates a bounded sample from an unbouded continuous Scipy distribution.

    Uses [inverse transform sampling](https://en.wikipedia.org/wiki/Inverse_transform_sampling)
    to draw samples from the unbounded distribution.

    The quantiles sampled uniformly are bounded, such that their transform is also implicitly
    bounded.

    Args:
        sampling_distribution (scipy.stats.rv_continuous): the unbouded continuous
            Scipy distribution
        bounds (tuple[float, float]): the bounds
        rng (RNG): the random number generator
        num_samples (int): the number of samples to draw

    Returns:
        jtyping.Float[np.ndarray, " num_samples"]: the samples from the bounded distribution
    """
    u = rng.uniform(low=0.0, high=1.0, size=(num_samples,))

    truncated_u = u * (
        sampling_distribution.cdf(bounds[1]) - sampling_distribution.cdf(bounds[0])
    ) + sampling_distribution.cdf(bounds[0])

    truncated_samples = sampling_distribution.ppf(q=truncated_u)

    return truncated_samples



from prob_conf_mat.experiment_aggregation.abc import (
    AGGREGATION_REGISTRY,
    ExperimentAggregator,
)
from prob_conf_mat.utils import RNG


def get_experiment_aggregator(
    aggregation: str, rng: RNG, **kwargs
) -> ExperimentAggregator:
    if aggregation not in AGGREGATION_REGISTRY:
        raise ValueError(
            f"Parameter `aggregation` must be a registered aggregation method. Currently: {aggregation}. Must be one of {set(AGGREGATION_REGISTRY.keys())}"
        )

    aggregator_instance = AGGREGATION_REGISTRY[aggregation](rng=rng, **kwargs)

    aggregator_instance._init_params = dict(aggregation=aggregation, **kwargs)

    return aggregator_instance



from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import jaxtyping as jtyping

    from prob_conf_mat.experiment_group import ExperimentGroup
    from prob_conf_mat.experiment import ExperimentResult
    from prob_conf_mat.experiment_aggregation.heterogeneity import HeterogeneityResult
    from prob_conf_mat.utils import RNG
    from prob_conf_mat.utils.typing import MetricLike

import inspect
import typing
from abc import ABCMeta, abstractmethod
from dataclasses import dataclass

import numpy as np

from prob_conf_mat.experiment_aggregation.heterogeneity import estimate_i2

AGGREGATION_REGISTRY = dict()


class ExperimentAggregator(metaclass=ABCMeta):
    """The abstract base class for experiment aggregation methods.

    Properties should be implemented as class attributes in derived metrics

    The `compute_metric` method needs to be implemented

    """

    def __init__(self, rng: RNG) -> None:
        self.rng = rng
        self._init_params = dict()

    def __init_subclass__(cls, **kwargs) -> None:
        super().__init_subclass__(**kwargs)

        # Validate =============================================================
        # Make sure that all aliases are unique
        for alias in cls.aliases:  # type: ignore
            if alias in AGGREGATION_REGISTRY:
                raise ValueError(
                    f"Alias '{alias}' not unique. Currently used by {AGGREGATION_REGISTRY[alias]}.",  # noqa: E501
                )

        # Register =============================================================
        for alias in cls.aliases:  # type: ignore
            AGGREGATION_REGISTRY[alias] = cls

        cls._kwargs = {
            param.name: param.annotation
            for param in inspect.signature(cls).parameters.values()
        }

    @property
    @abstractmethod
    def full_name(self) -> str:  # type: ignore
        """A human-readable name for this experiment-aggregation method."""
        raise NotImplementedError

    full_name: str

    @property
    @abstractmethod
    def aliases(self) -> list[str]:  # type: ignore
        """A list of all valid aliases for this metric. Can be used in configuration files."""
        raise NotImplementedError

    aliases: list[str]

    @property
    def name(self) -> str:
        """The name of the experiment aggregation method."""
        init_name = self._init_params.get("aggregation", None)
        if init_name is not None:
            return init_name
        return self.aliases[0]

    @abstractmethod
    def aggregate(
        self,
        experiment_samples: jtyping.Float[np.ndarray, " num_samples num_experiments"],
        bounds: tuple[float, float],
    ) -> jtyping.Float[np.ndarray, " num_experiments"]:
        """Aggregates samples from many experiments.

        Args:
            experiment_samples (Float[nd.array, "num_samples num_experiments"]): the samples from
                the individual experiments
            bounds (tuple[float, float]): the maximum and minimum possible value that the samples
                might take

        Returns:
            Float[nd.array, "num_samples num_experiments"]: samples from the aggregate distribution
        """
        raise NotImplementedError

    def _mappable_aggregate(self, kwargs: dict):
        return self.aggregate(**kwargs)

    def __call__(
        self,
        experiment_group: ExperimentGroup,
        metric: MetricLike,
        experiment_results: typing.Annotated[list[ExperimentResult], "num_experiments"],
    ) -> ExperimentAggregationResult:
        """Aggregates a series of experiment results from a specific ExperimentGroup and Metric."""
        # Stack the experiment values
        stacked_experiment_results: jtyping.Float[
            np.ndarray,
            " num_samples #num_classes num_experiments",
        ] = np.stack(
            [experiment_result.values for experiment_result in experiment_results],
            axis=-1,
        )

        # For each class, get the aggregated values
        all_class_aggregated_experiment_result = []
        all_class_heterogeneity = []
        for class_label in range(stacked_experiment_results.shape[1]):
            per_class_stacked_experiment_results: jtyping.Float[
                np.ndarray,
                " num_samples num_experiments",
            ] = stacked_experiment_results[:, class_label, :]

            per_class_aggregated_experiment_result: jtyping.Float[
                np.ndarray,
                " num_samples",
            ] = self.aggregate(
                experiment_samples=per_class_stacked_experiment_results,
                bounds=metric.bounds,  # type: ignore
            )

            all_class_aggregated_experiment_result.append(
                per_class_aggregated_experiment_result,
            )

            all_class_heterogeneity.append(
                estimate_i2(individual_samples=per_class_stacked_experiment_results),
            )

        # Finally, stack everything back together
        aggregated_experiment_result: jtyping.Float[
            np.ndarray,
            " num_samples #num_classes",
        ] = np.vstack(per_class_aggregated_experiment_result)  # type: ignore

        result = ExperimentAggregationResult(
            experiment_group=experiment_group,
            aggregator=self,  # type: ignore
            metric=metric,
            heterogeneity_results=all_class_heterogeneity,
            values=aggregated_experiment_result,
        )

        return result

    def __repr__(self) -> str:  # noqa: D105
        return f"ExperimentAggregator({self.name})"

    def __str__(self) -> str:  # noqa: D105
        return f"ExperimentAggregator({self.name})"

    def __eq__(self, other: object) -> bool:  # noqa: D105
        return isinstance(other, self.__class__) and self.name == other.name

    def __hash__(self):  # noqa: D105
        return hash(self.name)


@dataclass(frozen=True)
class ExperimentAggregationResult:
    """Class containing results from performing experiment aggregation."""

    experiment_group: ExperimentGroup
    aggregator: type[ExperimentAggregator]
    metric: MetricLike
    heterogeneity_results: list[HeterogeneityResult]
    values: jtyping.Float[np.ndarray, " num_samples #num_classes"]

    @property
    def name(self) -> str:
        """The name of the experiment group."""
        return self.experiment_group.name

    def __repr__(self) -> str:
        return (
            f"ExperimentAggregationResult(experiment_group={self.experiment_group}, "
            f"metric={self.metric}, aggregator={self.aggregator})"
        )

    def __str__(self) -> str:
        return (
            f"ExperimentAggregationResult(experiment_group={self.experiment_group}, "
            f"metric={self.metric}, aggregator={self.aggregator})"
        )



from __future__ import annotations
import typing

if typing.TYPE_CHECKING:
    import jaxtyping as jtyping

    from prob_conf_mat.utils.typing import MetricLike

from dataclasses import dataclass

import numpy as np

from prob_conf_mat.utils.formatting import fmt
from prob_conf_mat.stats import (
    summarize_posterior,
    PosteriorSummary,
    wilson_score_interval,
)

DELTA = "Δ"


@dataclass(frozen=True)
class PairwiseComparisonResult:
    lhs_name: str | None
    rhs_name: str | None
    metric: MetricLike

    observed_diff: float | None
    diff_dist: jtyping.Float[np.ndarray, " num_samples"]
    diff_dist_summary: PosteriorSummary

    direction: str
    p_direction: float
    p_direction_interpretation: str
    p_direction_score_interval_width: float

    min_sig_diff: float
    p_sig_neg: float
    p_rope: float
    p_sig_pos: float

    p_bi_sig: float
    p_bi_sig_interpretation: str
    p_bi_sig_score_interval_width: float

    p_rope_random: float
    bf_rope: float

    p_uni_sig: float
    # p_uni_sig_interpretation: str
    p_uni_sig_score_interval_width: float

    def template_sentence(self, precision: int = 4):
        # Build the template sentence
        template_sentence = ""
        template_sentence += f"Experiment {self.lhs_name}'s {self.metric.name} being"
        template_sentence += (
            " greater" if self.diff_dist_summary.median > 0 else " lesser"
        )
        template_sentence += f" than {self.rhs_name} could be considered"
        template_sentence += f" '{self.p_direction_interpretation}'* "

        # Existence statistics
        template_sentence += f"(Median {DELTA}={fmt(float(self.diff_dist_summary.median), precision=precision)}, "
        template_sentence += f"{fmt(self.diff_dist_summary.ci_probability, precision=precision, mode='%')} HDI="
        template_sentence += (
            f"[{fmt(self.diff_dist_summary.hdi[0], precision=4, mode='f')}, "
        )
        template_sentence += (
            f"{fmt(self.diff_dist_summary.hdi[1], precision=4, mode='f')}], "
        )
        template_sentence += (
            f"p_direction={fmt(self.p_direction, precision=4, mode='%')}).\n\n"
        )

        # Bidirectional significance
        template_sentence += (
            f"There is a {fmt(self.p_bi_sig, precision=precision, mode='%')}"
        )
        template_sentence += (
            " probability that this difference is bidirectionally significant"
        )
        template_sentence += (
            f" (ROPE=[{fmt(-self.min_sig_diff, precision=precision, mode='f')}, "
        )
        template_sentence += (
            f"{fmt(self.min_sig_diff, precision=precision, mode='f')}], "
        )
        template_sentence += (
            f"p_ROPE={fmt(self.p_rope, precision=precision, mode='%')}).\n\n"
        )
        template_sentence += (
            f"Bidirectional significance could be considered "
            f"'{self.p_bi_sig_interpretation}'*.\n\n"
        )

        # Unidirectional significance
        template_sentence += (
            f"There is a {fmt(self.p_uni_sig, precision=precision, mode='%')}"
        )
        template_sentence += (
            f" probability that this difference is significantly {self.direction}"
        )
        template_sentence += (
            f" (p_pos={fmt(self.p_sig_pos, precision=precision, mode='%')},"
        )
        template_sentence += (
            f" p_neg={fmt(self.p_sig_neg, precision=precision, mode='%')}).\n\n"
        )

        if self.bf_rope is not None:
            # Bidirectional significance to random
            template_sentence += (
                f"Relative to two random models "
                f"(p_ROPE,random={fmt(self.p_rope_random, precision=precision, mode='%')})"
            )

            log_bf_rope = np.log10(self.bf_rope)
            bf_rope_direction = "less" if np.sign(log_bf_rope) > 0.0 else "more"
            bf_rope_magnitude = np.power(10, np.abs(log_bf_rope))

            template_sentence += (
                f" significance is {fmt(bf_rope_magnitude, precision=precision, mode='f')} times "
                f"{bf_rope_direction} likely.\n\n"
            )

        template_sentence += (
            "* These interpretations are based off of loose guidelines, "
            "and should change according to the application."
        )

        return template_sentence


def pd_interpretation_guideline(
    pd: float,
) -> (
    typing.Literal["certain"]
    | typing.Literal["probable"]
    | typing.Literal["likely"]
    | typing.Literal["possible"]
    | typing.Literal["dubious"]
):
    # https://easystats.github.io/bayestestR/articles/guidelines.html#existence
    if pd < 0.0 or pd > 1.0:
        raise ValueError(f"Encountered pd value of {pd}, outside of range (0, 1).")
    if pd > 0.999:
        existence = "certain"
    elif pd > 0.99:
        existence = "probable"
    elif pd > 0.97:
        existence = "likely"
    elif pd > 0.95:
        existence = "possible"
    elif pd <= 0.95:
        existence = "dubious"
    else:
        raise ValueError(
            f"Encountered pd value of {pd}, somehow outside of range (0, 1).",
        )

    return existence


def p_rope_interpretation_guideline(
    p_rope: float,
) -> (
    typing.Literal["certain"]
    | typing.Literal["probable"]
    | typing.Literal["undecided"]
    | typing.Literal["probably negligible"]
    | typing.Literal["negligible"]
):
    # https://easystats.github.io/bayestestR/articles/guidelines.html#significance
    if p_rope < 0.0 or p_rope > 1.0:
        raise ValueError(f"Encountered p_rope value of {p_rope}, outside of range.")
    if p_rope < 0.01:
        significance = "certain"
    elif p_rope < 0.025:
        significance = "probable"
    elif p_rope >= 0.025 and p_rope <= 0.975:
        significance = "undecided"
    elif p_rope > 0.975 and p_rope <= 0.99:
        significance = "probably negligible"
    elif p_rope > 0.99:
        significance = "negligible"
    else:
        raise ValueError(
            f"Encountered p_rope value of {p_rope}, somehow outside of range.",
        )

    return significance


def pairwise_compare(
    metric: MetricLike,
    diff_dist: jtyping.Float[np.ndarray, " num_samples"],
    ci_probability: float,
    min_sig_diff: float | None = None,  # type: ignore
    lhs_name: str | None = None,
    rhs_name: str | None = None,
    random_diff_dist: jtyping.Float[np.ndarray, " num_samples"] | None = None,
    observed_difference: float | None = None,
) -> PairwiseComparisonResult:
    # Find central tendency of diff dit
    diff_dist_summary = summarize_posterior(diff_dist, ci_probability=ci_probability)

    # Probability of existence
    if diff_dist_summary.median > 0:
        pd: float = np.mean(diff_dist > 0)  # type: ignore
    else:
        pd: float = np.mean(diff_dist < 0)  # type: ignore

    pd_interpretation = pd_interpretation_guideline(pd=pd)

    # Define a default ROPE
    if min_sig_diff is None:
        min_sig_diff: float = 0.1 * np.std(diff_dist)  # type: ignore

    # Count the number of instances within each bin
    # Significantly negative, within ROPE, significantly positive
    counts, _ = np.histogram(
        diff_dist,
        bins=[-float("inf"), -min_sig_diff, min_sig_diff + 1e-8, float("inf")],
    )

    p_sig_neg, p_rope, p_sig_pos = counts / diff_dist.shape[0]

    p_bi_sig = 1 - p_rope
    p_rope_interpretation = p_rope_interpretation_guideline(p_rope)

    # Compare p_ROPE to random distributions
    if random_diff_dist is not None and min_sig_diff > 0.0:
        p_bi_sig_random = np.mean(
            (random_diff_dist < -min_sig_diff) | (random_diff_dist > min_sig_diff),
        )
        p_rope_random = 1 - p_bi_sig_random
        bf_rope = p_rope / p_rope_random

    else:
        p_bi_sig_random = None
        p_rope_random = None
        bf_rope = None

    result = PairwiseComparisonResult(
        # Admin
        lhs_name=lhs_name,
        rhs_name=rhs_name,
        metric=metric,
        # The difference distribution
        observed_diff=observed_difference,
        diff_dist=diff_dist,
        diff_dist_summary=diff_dist_summary,
        # Existence
        direction="positive" if diff_dist_summary.median > 0 else "negative",
        p_direction=pd,
        p_direction_interpretation=pd_interpretation,
        p_direction_score_interval_width=wilson_score_interval(
            p=pd,
            n=diff_dist.shape[0],
        ),
        # Significance buckets
        min_sig_diff=min_sig_diff,
        p_sig_neg=p_sig_neg,
        p_rope=p_rope,
        p_sig_pos=p_sig_pos,
        # Bidirectional significance
        p_bi_sig=p_bi_sig,
        p_bi_sig_interpretation=p_rope_interpretation,
        p_bi_sig_score_interval_width=wilson_score_interval(
            p=1 - p_rope,
            n=diff_dist.shape[0],
        ),
        # Significance relative to random
        p_rope_random=p_rope_random,  # type: ignore
        bf_rope=bf_rope,  # type: ignore
        # Unidirectional significance
        p_uni_sig=p_sig_pos if diff_dist_summary.median > 0 else p_sig_neg,
        p_uni_sig_score_interval_width=wilson_score_interval(
            p=p_sig_pos if diff_dist_summary.median > 0 else p_sig_neg,
            n=diff_dist.shape[0],
        ),
    )

    return result



from __future__ import annotations
import typing

if typing.TYPE_CHECKING:
    import jaxtyping as jtyping

    from prob_conf_mat.utils import RNG

import numpy as np


_DIRICHLET_PRIOR_STRATEGIES = {
    "bayes-laplace": 1.0,
    "bayes": 1.0,
    "laplace": 1.0,
    "ones": 1.0,
    "one": 1.0,
    "jeffrey": 0.5,
    "jeffreys": 0.5,
    "halves": 0.5,
    "half": 0.5,
    "haldane": 0.0,
    "zeros": 0.0,
    "zero": 0.0,
}


# TODO: add support for other dtypes
def dirichlet_prior(
    strategy: str | float | int | jtyping.Float[np.typing.ArrayLike, " ..."],
    shape: tuple[int, ...],
) -> jtyping.Float[np.ndarray, " ..."]:
    """Creates a prior array for a Dirichlet distribution.

    Returns:
        jtyping.Float[np.ndarray, " ..."]: the prior vector
    """
    if isinstance(strategy, float | int):
        if strategy < 0:
            raise ValueError(
                f"A Dirichlet prior must contain positive values. Received: {strategy}",
            )

        prior = np.full(shape, fill_value=strategy, dtype=np.float64)

    elif isinstance(strategy, str):
        if strategy not in _DIRICHLET_PRIOR_STRATEGIES:
            raise ValueError(
                f"Prior strategy `{strategy}` not recognized. "
                f"Choose one of: {set(_DIRICHLET_PRIOR_STRATEGIES.keys())}",
            )

        strategy_fill_value = _DIRICHLET_PRIOR_STRATEGIES[strategy]
        prior = np.full(shape, fill_value=strategy_fill_value, dtype=np.float64)

    else:
        try:
            # TODO: control dtype
            prior = np.array(strategy, dtype=np.float64)
        except Exception as e:  # noqa: BLE001
            raise ValueError(
                f"While trying to convert {strategy} to a numpy array, "
                f"received the following error:\n{e}",
            )

        if prior.shape != shape:
            raise ValueError(
                f"Prior does not match required shape, {prior.shape} != {shape}. "
                f"Parsed {prior} of type {type(prior)} from {strategy} fo type {type(strategy)}.",
            )

    return prior


def dirichlet_sample(
    rng: RNG,
    alphas: jtyping.Float[np.ndarray, " ..."],
    num_samples: int,
) -> jtyping.Float[np.ndarray, " num_samples ..."]:
    """Generate Dirichlet distributed samples from an array of Gamma distributions.

    A Dirichlet distribution can be constructed by [dividing a set of Gamma distributions by their sum](https://en.wikipedia.org/wiki/Dirichlet_distribution#Related_distributions).

    For some reason the Numpy implementation of the [Dirichlet distribution](https://numpy.org/doc/stable/reference/random/generated/numpy.random.Generator.dirichlet.html#numpy.random.Generator.dirichlet)
    is not vectorized, while the implementation of the [Gamma distribution](https://numpy.org/doc/stable/reference/random/generated/numpy.random.Generator.gamma.html).
    is.

    Adapted from [StackOverflow](https://stackoverflow.com/a/15917312).

    This function is the performance bottleneck for this package.
    Need to make sure it's performant.

    Args:
        rng (RNG): the random number generator
        alphas (jtyping.Float[np.ndarray, "..."]): the Dirichlet parameters
        num_samples (int): the number of samples to retrieve

    Returns:
        jtyping.Float[np.ndarray, " num_samples ..."]: samples from the specified Dirichlet distribution
    """  # noqa: E501
    # TODO: add support for other dtypes
    alphas = alphas.astype(np.float64)

    # Broadcast alphas to the desired shape
    alphas = np.broadcast_to(alphas, (num_samples, *alphas.shape))

    # Generate independent gamma-distributed samples
    # This bit dominates the run-time
    arr = rng.standard_gamma(alphas)

    # Normalize the sampled gamma distributed variables
    # This bit is probably as fast as it can get
    np.divide(arr, np.einsum("...i->...", arr)[..., np.newaxis], out=arr)

    return arr



import jaxtyping as jtyping
import numpy as np

from prob_conf_mat.metrics.abc import Metric


class PrecisionGain(Metric):
    # TODO: write documentation

    full_name = "Precision Gain"
    is_multiclass = False
    bounds = (0.0, 1.0)
    dependencies = ("precision", "prevalence")
    sklearn_equivalent = None
    aliases = ["precision_gain"]

    def compute_metric(
        self,
        precision: jtyping.Float[np.ndarray, "num_samples num_classes"],
        prevalence: jtyping.Float[np.ndarray, "num_samples num_classes"],
    ) -> jtyping.Float[np.ndarray, " num_samples num_classes num_classes"]:
        precision_gain = (precision - prevalence) / ((1 - prevalence) * precision)

        precision_gain = np.clip(precision_gain, a_min=0.0, a_max=1.0)

        return precision_gain


class RecallGain(Metric):
    # TODO: write documentation

    full_name = "Recall Gain"
    is_multiclass = False
    bounds = (0.0, 1.0)
    dependencies = ("recall", "prevalence")
    sklearn_equivalent = None
    aliases = ["recall_gain"]

    def compute_metric(
        self,
        recall: jtyping.Float[np.ndarray, "num_samples num_classes"],
        prevalence: jtyping.Float[np.ndarray, "num_samples num_classes"],
    ) -> jtyping.Float[np.ndarray, " num_samples num_classes num_classes"]:
        recall_gain = (recall - prevalence) / ((1 - prevalence) * recall)

        recall_gain = np.clip(recall_gain, a_min=0.0, a_max=1.0)

        return recall_gain


class F1Gain(Metric):
    # TODO: write documentation

    full_name = "F1 Gain"
    is_multiclass = False
    bounds = (0.0, 1.0)
    dependencies = ("f1", "prevalence")
    sklearn_equivalent = None
    aliases = ["f1_gain"]

    def compute_metric(
        self,
        f1: jtyping.Float[np.ndarray, "num_samples num_classes"],
        prevalence: jtyping.Float[np.ndarray, "num_samples num_classes"],
    ) -> jtyping.Float[np.ndarray, " num_samples num_classes num_classes"]:
        f1_gain = (f1 - prevalence) / ((1 - prevalence) * f1)

        f1_gain = np.clip(f1_gain, a_min=0.0, a_max=1.0)

        return f1_gain


class FBetaGain(Metric):
    # TODO: write documentation

    full_name = "FBeta Gain"
    is_multiclass = False
    bounds = (0.0, 1.0)
    dependencies = (
        "ppv",
        "tpr",
        "prevalence",
    )
    sklearn_equivalent = None
    aliases = ["fbeta_gain"]

    def __init__(self, beta: float = 1.0):
        super().__init__()

        self.beta = beta

    def compute_metric(
        self,
        ppv: jtyping.Float[np.ndarray, "num_samples num_classes"],
        tpr: jtyping.Float[np.ndarray, "num_samples num_classes"],
        prevalence: jtyping.Float[np.ndarray, "num_samples num_classes"],
    ) -> jtyping.Float[np.ndarray, " num_samples num_classes num_classes"]:
        beta_2 = self.beta**2

        fbeta = (1 + beta_2) * (ppv * tpr) / (beta_2 * ppv + tpr)

        # In case one of the ratios is nan (most likely due to 0 division), set to 0
        fbeta = np.nan_to_num(
            fbeta,
            nan=0.0,
        )

        fbeta_gain = (fbeta - prevalence) / ((1 - prevalence) * fbeta)

        fbeta_gain = np.clip(fbeta_gain, a_min=0.0, a_max=1.0)

        return fbeta_gain



import typing

import pytest
import numpy as np

from prob_conf_mat import Study
from prob_conf_mat.config import ConfigError, ConfigWarning


class TestConfig:
    base_config = dict(
        seed=0,
        num_samples=10000,
        ci_probability=0.95,
    )

    def fetch_base_config(self, *args) -> dict[str, typing.Any]:
        return {k: v for k, v in self.base_config.items() if k not in args}

    def test_seed(self) -> None:
        # Test a valid seed
        Study(seed=0, **self.fetch_base_config("seed"))

        # Test a negative int
        with pytest.raises(
            ConfigError,
            match="Parameter `seed` must be a positive integer.",
        ):
            Study(seed=-1, **self.fetch_base_config("seed"))

        # Test a non-convertible non-int class
        with pytest.raises(
            TypeError,
            match="Parameter `seed` must be a positive integer",
        ):
            Study(seed="foo", **self.fetch_base_config("seed"))  # type: ignore

        # Test a convertible non-int class
        with pytest.warns(
            ConfigWarning,
            match="Parameter `seed` must be a positive integer",
        ):
            Study(seed=0.1, **self.fetch_base_config("seed"))  # type: ignore

        # Test a negative convertible non-int class
        with (
            pytest.warns(
                ConfigWarning,
                match="Parameter `seed` must be a positive integer",
            ),
            pytest.raises(
                ConfigError,
                match="Parameter `seed` must be a positive integer.",
            ),
        ):
            Study(seed=-10.1, **self.fetch_base_config("seed"))  # type: ignore

        # Test a 'None'
        with pytest.warns(
            ConfigWarning,
            match="Recieved `None` as seed. Defaulting to fractional seconds:",
        ):
            Study(seed=None, **self.fetch_base_config("seed"))

    def test_num_samples(self) -> None:
        # Test a valid num_samples
        Study(num_samples=10000, **self.fetch_base_config("num_samples"))

        # Test a low value
        with pytest.warns(
            ConfigWarning,
            match="Parameter `num_samples` should be large to reduce variability",
        ):
            Study(num_samples=1, **self.fetch_base_config("num_samples"))

        # Test a negative int
        with pytest.raises(
            ConfigError,
            match="Parameter `num_samples` must be greater than 0. Currently:",
        ):
            Study(num_samples=-10000, **self.fetch_base_config("num_samples"))

        # Test a None
        with pytest.warns(
            ConfigWarning,
            match="Parameter `num_samples` is `None`. Setting to default value of 10000.",
        ):
            Study(num_samples=None, **self.fetch_base_config("num_samples"))

        # Test a non-convertible non-int class
        with pytest.raises(
            TypeError,
            match="Parameter `num_samples` must be a strictly positive integer",
        ):
            Study(num_samples="foo", **self.fetch_base_config("num_samples"))  # type: ignore

        # Test a convertible non-int class
        with pytest.warns(
            ConfigWarning,
            match="Parameter `num_samples` must be a strictly positive integer",
        ):
            Study(num_samples=10e5, **self.fetch_base_config("num_samples"))  # type: ignore

        # Test a negative convertible non-int class
        with (
            pytest.warns(
                ConfigWarning,
                match="Parameter `num_samples` must be a strictly positive integer",
            ),
            pytest.raises(
                ConfigError,
                match="Parameter `num_samples` must be greater than 0. Currently:",
            ),
        ):
            Study(num_samples=(-10e5), **self.fetch_base_config("num_samples"))  # type: ignore

    def test_ci_probability(self) -> None:
        # Test a valid num_samples
        Study(ci_probability=0.95, **self.fetch_base_config("ci_probability"))

        # Test CI probability bounds
        with pytest.raises(
            ConfigError,
            match=r"Parameter `ci_probability` must be within \(0.0, 1.0\]",
        ):
            Study(ci_probability=0.0, **self.fetch_base_config("ci_probability"))

        with pytest.raises(
            ConfigError,
            match=r"Parameter `ci_probability` must be within \(0.0, 1.0\]",
        ):
            Study(ci_probability=-0.5, **self.fetch_base_config("ci_probability"))

        with pytest.raises(
            ConfigError,
            match=r"Parameter `ci_probability` must be within \(0.0, 1.0\]",
        ):
            Study(ci_probability=1.5, **self.fetch_base_config("ci_probability"))

        # Test a None
        with pytest.warns(
            ConfigWarning,
            match="Parameter `ci_probability` is `None`. Setting to default value of 0.95",
        ):
            Study(ci_probability=None, **self.fetch_base_config("ci_probability"))

        # Test a non-convertible non-float class
        with pytest.raises(
            TypeError,
            match="Parameter `ci_probability` must be a float",
        ):
            Study(ci_probability="foo", **self.fetch_base_config("ci_probability"))  # type: ignore

        with pytest.raises(
            TypeError,
            match="Parameter `ci_probability` must be a float",
        ):
            Study(
                ci_probability=np.array([[0.95], [0.95]]),  # type: ignore
                **self.fetch_base_config("ci_probability"),
            )

        # Test a convertible non-float class
        with pytest.warns(
            ConfigWarning,
            match="Parameter `ci_probability` must be a float",
        ):
            Study(
                ci_probability=np.array(0.95),  # type: ignore
                **self.fetch_base_config("ci_probability"),
            )  # type: ignore

            with pytest.warns(
                DeprecationWarning,
                match="Conversion of an array with ndim > 0 to a scalar is deprecated",
            ):
                Study(
                    ci_probability=np.array([0.95]),  # type: ignore
                    **self.fetch_base_config("ci_probability"),
                )

    def test_prevalence_prior(self) -> None:
        # ======================================================================
        # Prevalence prior
        # ======================================================================
        # Test default prior
        with pytest.warns(
            ConfigWarning,
            match=r"Defaulting to the 0 \(Haldane\) prior\.",
        ):
            study = Study(**self.fetch_base_config())

            study.add_experiment(
                experiment_name="test/test",
                confusion_matrix=np.array([[1, 0], [0, 1]]),
                confusion_prior=0,
            )

        # Test None as prior
        with pytest.warns(
            ConfigWarning,
            match=r"Defaulting to the 0 \(Haldane\) prior\.",
        ):
            study = Study(**self.fetch_base_config())

            study.add_experiment(
                experiment_name="test/test",
                confusion_matrix=np.array([[1, 0], [0, 1]]),
                prevalence_prior=None,
                confusion_prior=0,
            )

        # Test numeric as prior
        # Integer
        study = Study(**self.fetch_base_config())

        study.add_experiment(
            experiment_name="test/test",
            confusion_matrix=np.array([[1, 0], [0, 1]]),
            prevalence_prior=1,
            confusion_prior=0,
        )

        assert np.all(
            study["test/test"].prevalence_prior == 1.0,  # type: ignore
        )

        # Float
        study = Study(**self.fetch_base_config())

        study.add_experiment(
            experiment_name="test/test",
            confusion_matrix=np.array([[1, 0], [0, 1]]),
            prevalence_prior=1.0,
            confusion_prior=0,
        )

        assert np.all(
            a=study["test/test"].prevalence_prior  # type: ignore
            == 1.0,
        )

        # Out of bounds numeric
        with pytest.raises(
            ConfigError,
            match=r"If numeric, must be greater than 0",
        ):
            study = Study(**self.fetch_base_config())

            study.add_experiment(
                experiment_name="test/test",
                confusion_matrix=np.array([[1, 0], [0, 1]]),
                prevalence_prior=-1.0,
                confusion_prior=0,
            )

        # Arraylike ============================================================
        # Direct
        study = Study(**self.fetch_base_config())

        study.add_experiment(
            experiment_name="test/test",
            confusion_matrix=np.array([[1, 0], [0, 1]]),
            prevalence_prior=[1.0, 1.0],
            confusion_prior=0,
        )

        assert np.all(
            a=study["test/test"].prevalence_prior  # type: ignore
            == 1.0,
        )

        # Needs reshaping
        study = Study(**self.fetch_base_config())

        study.add_experiment(
            experiment_name="test/test",
            confusion_matrix=np.array([[1, 0], [0, 1]]),
            prevalence_prior=[[1.0, 1.0]],
            confusion_prior=0,
        )

        assert np.all(
            a=study["test/test"].prevalence_prior  # type: ignore
            == 1.0,
        )

        # Malformed - too many columns
        with pytest.raises(
            ConfigError,
            match=r"prevalence prior is malformed",
        ):
            study = Study(**self.fetch_base_config())

            study.add_experiment(
                experiment_name="test/test",
                confusion_matrix=np.array([[1, 0], [0, 1]]),
                prevalence_prior=[[1.0, 1.0, 1.0]],
                confusion_prior=0,
            )

        # Malformed - too many rows
        with pytest.raises(
            ConfigError,
            match=r"prevalence prior is malformed",
        ):
            study = Study(**self.fetch_base_config())

            study.add_experiment(
                experiment_name="test/test",
                confusion_matrix=np.array([[1, 0], [0, 1]]),
                prevalence_prior=[[1.0, 1.0], [1.0, 1.0]],
                confusion_prior=0,
            )

        # Malformed - too few columns
        with pytest.raises(
            ConfigError,
            match=r"prevalence prior is malformed",
        ):
            study = Study(**self.fetch_base_config())

            study.add_experiment(
                experiment_name="test/test",
                confusion_matrix=np.array([[1, 0], [0, 1]]),
                prevalence_prior=[
                    [
                        1.0,
                    ],
                ],
                confusion_prior=0,
            )

        # Malformed - too few rows
        with pytest.raises(
            ConfigError,
            match=r"prevalence prior is malformed",
        ):
            study = Study(**self.fetch_base_config())

            study.add_experiment(
                experiment_name="test/test",
                confusion_matrix=np.array([[1, 0], [0, 1]]),
                prevalence_prior=[],
                confusion_prior=0,
            )

        # Invalid values in an arraylike
        with pytest.raises(
            ConfigError,
            match=r"all values must be positive",
        ):
            study = Study(**self.fetch_base_config())

            study.add_experiment(
                experiment_name="test/test",
                confusion_matrix=np.array([[1, 0], [0, 1]]),
                prevalence_prior=[-1.0, 1.0],
                confusion_prior=0,
            )

        with pytest.raises(
            ConfigError,
            match=r"prevalence prior is invalid",
        ):
            study = Study(**self.fetch_base_config())

            study.add_experiment(
                experiment_name="test/test",
                confusion_matrix=np.array([[1, 0], [0, 1]]),
                prevalence_prior=["one", 1.0],
                confusion_prior=0,
            )

    def test_confusion_matrix(self) -> None:
        # ======================================================================
        # Confusion prior
        # ======================================================================
        # Test default prior
        with pytest.warns(
            ConfigWarning,
            match=r"Defaulting to the 0 \(Haldane\) prior\.",
        ):
            study = Study(**self.fetch_base_config())

            study.add_experiment(
                experiment_name="test/test",
                confusion_matrix=np.array([[1, 0], [0, 1]]),
                prevalence_prior=0,
            )

        # Test None as prior
        with pytest.warns(
            ConfigWarning,
            match=r"Defaulting to the 0 \(Haldane\) prior\.",
        ):
            study = Study(**self.fetch_base_config())

            study.add_experiment(
                experiment_name="test/test",
                confusion_matrix=np.array([[1, 0], [0, 1]]),
                prevalence_prior=0,
                confusion_prior=None,
            )

        # Test numeric as prior
        # Integer
        study = Study(**self.fetch_base_config())

        study.add_experiment(
            experiment_name="test/test",
            confusion_matrix=np.array([[1, 0], [0, 1]]),
            prevalence_prior=0,
            confusion_prior=1,
        )

        assert np.all(
            study["test/test"].confusion_prior == 1.0,  # type: ignore
        )

        # Float
        study = Study(**self.fetch_base_config())

        study.add_experiment(
            experiment_name="test/test",
            confusion_matrix=np.array([[1, 0], [0, 1]]),
            prevalence_prior=0,
            confusion_prior=1.0,
        )

        assert np.all(
            a=study["test/test"].confusion_prior == 1.0,  # type: ignore
        )

        # Out of bounds numeric
        with pytest.raises(
            ConfigError,
            match=r"If numeric, must be greater than 0",
        ):
            study = Study(**self.fetch_base_config())

            study.add_experiment(
                experiment_name="test/test",
                confusion_matrix=np.array([[1, 0], [0, 1]]),
                prevalence_prior=0,
                confusion_prior=-1.0,
            )

        # Arraylike ============================================================
        # Direct
        study = Study(**self.fetch_base_config())

        study.add_experiment(
            experiment_name="test/test",
            confusion_matrix=np.array([[1, 0], [0, 1]]),
            prevalence_prior=0,
            confusion_prior=[[1.0, 1.0], [1.0, 1.0]],
        )

        assert np.all(
            a=study["test/test"].confusion_prior == 1.0,  # type: ignore
        )

        # Needs reshaping
        study = Study(**self.fetch_base_config())

        study.add_experiment(
            experiment_name="test/test",
            confusion_matrix=np.array([[1, 0], [0, 1]]),
            prevalence_prior=0,
            confusion_prior=[[[1.0, 1.0], [1.0, 1.0]]],
        )

        assert np.all(
            a=study["test/test"].confusion_prior == 1.0,  # type: ignore
        )

        # Malformed - too many columns
        with pytest.raises(
            ConfigError,
            match=r"confusion prior is malformed",
        ):
            study = Study(**self.fetch_base_config())

            study.add_experiment(
                experiment_name="test/test",
                confusion_matrix=np.array([[1, 0], [0, 1]]),
                prevalence_prior=0,
                confusion_prior=[[1.0, 1.0, 1.0], [1.0, 1.0, 1.0]],
            )

        # Malformed - too many rows
        with pytest.raises(
            ConfigError,
            match=r"confusion prior is malformed",
        ):
            study = Study(**self.fetch_base_config())

            study.add_experiment(
                experiment_name="test/test",
                confusion_matrix=np.array([[1, 0], [0, 1]]),
                prevalence_prior=0,
                confusion_prior=[[1.0, 1.0], [1.0, 1.0], [1.0, 1.0]],
            )

        # Malformed - too few columns
        with pytest.raises(
            ConfigError,
            match=r"confusion prior is malformed",
        ):
            study = Study(**self.fetch_base_config())

            study.add_experiment(
                experiment_name="test/test",
                confusion_matrix=np.array([[1, 0], [0, 1]]),
                prevalence_prior=0,
                confusion_prior=[
                    [
                        1.0,
                    ],
                    [
                        1.0,
                    ],
                ],
            )

        # Malformed - too few rows
        with pytest.raises(
            ConfigError,
            match=r"confusion prior is malformed",
        ):
            study = Study(**self.fetch_base_config())

            study.add_experiment(
                experiment_name="test/test",
                confusion_matrix=np.array([[1, 0], [0, 1]]),
                prevalence_prior=0,
                confusion_prior=[[1.0, 1.0]],
            )

        # Invalid values in an arraylike
        with pytest.raises(
            ConfigError,
            match=r"all values must be positive",
        ):
            study = Study(**self.fetch_base_config())

            study.add_experiment(
                experiment_name="test/test",
                confusion_matrix=np.array([[1, 0], [0, 1]]),
                prevalence_prior=0,
                confusion_prior=[[-1.0, 1.0], [1.0, 1.0]],
            )

        with pytest.raises(
            ConfigError,
            match=r"confusion prior is invalid",
        ):
            study = Study(**self.fetch_base_config())

            study.add_experiment(
                experiment_name="test/test",
                confusion_matrix=np.array([[1, 0], [0, 1]]),
                prevalence_prior=0,
                confusion_prior=[["one", 1.0], [1.0, 1.0]],
            )

        # TODO: multiple confusion matrices with different num_classes

    def test_adding_experiments(self) -> None:
        study = Study(**self.fetch_base_config())

        study.add_experiment(
            experiment_name="test/test",
            confusion_matrix=np.array([[1, 0], [0, 1]]),
            prevalence_prior=0,
            confusion_prior=0,
        )

        with pytest.warns(
            UserWarning,
            match=r"Experiment .* already exists. Overwriting",
        ):
            study = Study(**self.fetch_base_config())

            study.add_experiment(
                experiment_name="test/test_1",
                confusion_matrix=np.array([[1, 0], [0, 1]]),
                prevalence_prior=0,
                confusion_prior=0,
            )

            study.add_experiment(
                experiment_name="test/test_1",
                confusion_matrix=np.array([[1, 0], [0, 1]]),
                prevalence_prior=0,
                confusion_prior=0,
            )

        with pytest.raises(
            ConfigError,
            match=r"Experiment group .* has incongruent confusion matrices",
        ):
            study = Study(**self.fetch_base_config())

            study.add_experiment(
                experiment_name="test/test_1",
                confusion_matrix=np.array([[1, 0], [0, 1]]),
                prevalence_prior=0,
                confusion_prior=0,
            )

            study.add_experiment(
                experiment_name="test/test_2",
                confusion_matrix=np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]]),
                prevalence_prior=0,
                confusion_prior=0,
            )



import re
import logging
import sys
import typing
from dataclasses import dataclass, field
from collections import OrderedDict
from pathlib import Path

from tabulate import tabulate

from prob_conf_mat.utils import fmt
from prob_conf_mat.metrics import METRIC_REGISTRY, AVERAGING_REGISTRY
from prob_conf_mat.experiment_aggregation import AGGREGATION_REGISTRY

logger = logging.getLogger(__name__)

DOCUMENTATION_DIR = Path("documentation")

REFERENCE_PART = DOCUMENTATION_DIR / "Reference"

METRICS_AND_AVERAGING_CHAPTER = REFERENCE_PART / "Metrics"
METRICS_AND_AVERAGING_OVERVIEW = METRICS_AND_AVERAGING_CHAPTER / "index.md"
METRICS_SECTION = METRICS_AND_AVERAGING_CHAPTER / "Metrics.md"
AVERAGING_SECTION = METRICS_AND_AVERAGING_CHAPTER / "Averaging.md"

EXPERIMENT_AGGREGATION_CHAPTER = REFERENCE_PART / "Experiment Aggregation"
EXPERIMENT_AGGREGATION_SECTION = EXPERIMENT_AGGREGATION_CHAPTER / "index.md"
HETEROGENEITY_SECTION = EXPERIMENT_AGGREGATION_CHAPTER / "heterogeneity.md"

REPL_STRING = re.compile(r"@@(.+?)@@")
TEMPLATE_DIR = Path(__file__).parent


@dataclass
class KwargMatch:
    key: str | None = None
    spans: list[tuple[int, int]] = field(default_factory=lambda: [])
    value: typing.Any | None = None

    def __add__(self, other):
        if self.key is None:
            self.key = other.key

        self.spans = self.spans + other.spans

        return self


class Template:
    def __init__(self, file_name: str | Path) -> None:
        self.file_path: Path = TEMPLATE_DIR / file_name
        self.name = self.file_path.stem

        self.template = self.file_path.read_text()

        self.kwargs = OrderedDict()
        for match in REPL_STRING.finditer(self.template):
            self.kwargs[match.group(1)] = self.kwargs.get(
                match.group(1),
                KwargMatch(),
            ) + KwargMatch(key=match.group(1), spans=[match.span()])

    def set(self, key: str, value):
        self.kwargs[key].value = value

    def _format_value(self, value) -> str:
        if isinstance(value, str):
            return value
        if isinstance(value, float):
            return fmt(value)
        if isinstance(value, int):
            return f"{value:d}"
        try:
            return str(value)
        except Exception as e:
            raise TypeError(
                f"Unable to format {value} of type {type(value)}. Raises {e}",
            )

    def __repr__(self) -> str:
        return f"Template({self.name})"

    def __str__(self) -> str:
        filled_template = ""

        unsorted_spans = [
            (span, kwarg_match.key, kwarg_match.value)
            for kwarg_match in self.kwargs.values()
            for span in kwarg_match.spans
        ]
        sorted_spans = sorted(unsorted_spans, key=lambda x: x[0][0])

        left_pointer = 0
        for (begin, end), key, value in sorted_spans:
            filled_template += self.template[left_pointer:begin]

            if value is not None:
                filled_template += self._format_value(value)
            else:
                filled_template += f"@@{key}@@"

            left_pointer = end

        filled_template += self.template[left_pointer:]

        return filled_template


def metrics_and_averaging_overview() -> None:
    # Load in the template
    template_fp = TEMPLATE_DIR / "metrics_index.md"
    if not template_fp.exists():
        raise FileNotFoundError(f"Could not find a template file at '{template_fp}'")

    logger.info(f"Metrics & Averaging - Found template at '{template_fp}'")

    template = Template(
        file_name=template_fp,
    )

    # Complete the template
    # Metrics Table ============================================================
    # Generate a record for each metric alias
    aliases = sorted(METRIC_REGISTRY.items(), key=lambda x: x[0])
    aliases_index = []
    for i, (alias, metric) in enumerate(aliases):
        aliases_index += [
            [
                f"'{alias}'",
                f"[`{metric.__name__}`](Metrics.md#{metric.__module__}.{metric.__name__})",
                metric.is_multiclass,
                # metric.bounds,
                metric.sklearn_equivalent,
            ],
        ]

    # Creates a table with some important information as an overview
    template.set(
        "metrics_table",
        value=tabulate(
            tabular_data=aliases_index,
            headers=[
                "Alias",
                "Metric",
                "Multiclass",
                # "Bounds",
                "sklearn",
                "Tested",
            ],
            tablefmt="github",
        ),
    )

    logger.info(
        "Metrics & Averaging - Filled parameter `metrics_table`",
    )

    # Averaging list ===========================================================
    # Generate a record for each averaging alias
    aliases = sorted(AVERAGING_REGISTRY.items(), key=lambda x: x[0])
    aliases_index = []
    for i, (alias, avg_method) in enumerate(aliases):
        aliases_index += [
            [
                f"'{alias}'",
                # TODO: check that this works when hosting as well
                f"[`{avg_method.__name__}`](Averaging.md#{avg_method.__module__}.{avg_method.__name__})",
                avg_method.sklearn_equivalent,
            ],
        ]

    # Creates a table with some important information as an overview
    template.set(
        "averaging_table",
        value=tabulate(
            tabular_data=aliases_index,
            headers=[
                "Alias",
                "Metric",
                "sklearn",
            ],
            tablefmt="github",
        ),
    )

    logger.info(
        "Metrics & Averaging - Filled parameter `averaging_table`",
    )

    # Write the template to a md file
    output_fp = METRICS_AND_AVERAGING_OVERVIEW.resolve()
    output_fp.write_text(
        str(template),
        encoding="utf-8",
    )

    logger.info(
        f"Metrics & Averaging - Wrote filled template to '{output_fp}'",
    )


def metrics() -> None:
    # Load in the template
    template_fp = TEMPLATE_DIR / "metrics.md"
    if not template_fp.exists():
        raise FileNotFoundError(f"Could not find a template file at '{template_fp}'")

    logger.info(f"Metrics - Found template at '{template_fp}'")

    template = Template(
        file_name=template_fp,
    )

    # Complete the template
    # Metrics list =============================================================
    all_metrics = {str(metric): metric for metric in METRIC_REGISTRY.values()}

    template.set(
        "metrics_list",
        value="\n\n".join(
            (
                f"::: {metric.__module__}.{metric.__name__}"
                "\n    options:"
                "\n        heading_level: 3"
                "\n        show_root_heading: true"
                "\n        show_root_toc_entry: true"
                "\n        show_category_heading: false"
                "\n        show_symbol_type_toc: true"
                "\n        summary:"
                "\n                attributes: false"
                "\n                functions: false"
                "\n                modules: false"
                "\n        members:"
                "\n                - aliases"
                "\n                - bounds"
                "\n                - dependencies"
                "\n                - is_multiclass"
                "\n                - sklearn_equivalent"
                "\n        show_labels: false"
                "\n        group_by_category: false"
            )
            for metric in all_metrics.values()
        ),
    )

    logger.info(
        "Metrics - Filled parameter `metrics_list`",
    )

    # Write the template to a md file
    output_fp = METRICS_SECTION.resolve()
    output_fp.write_text(
        str(template),
        encoding="utf-8",
    )

    logger.info(
        f"Metrics - Wrote filled template to '{output_fp}'",
    )


def averaging() -> None:
    # Load in the template
    template_fp = TEMPLATE_DIR / "averaging.md"
    if not template_fp.exists():
        raise FileNotFoundError(f"Could not find a template file at '{template_fp}'")

    logger.info(f"Averaging - Found template at '{template_fp}'")

    template = Template(
        file_name=template_fp,
    )

    # Complete the template
    # Averaging methods list ===================================================
    all_avg_methods = {
        str(avg_method): avg_method for avg_method in AVERAGING_REGISTRY.values()
    }

    template.set(
        key="averaging_methods_list",
        value="\n\n".join(
            (
                f"::: {avg_method.__module__}.{avg_method.__name__}"
                "\n    options:"
                "\n        heading_level: 3"
                "\n        show_root_heading: true"
                "\n        show_root_toc_entry: true"
                "\n        show_category_heading: false"
                "\n        show_symbol_type_toc: true"
                "\n        summary:"
                "\n                attributes: false"
                "\n                functions: false"
                "\n                modules: false"
                "\n        members:"
                "\n                - aliases"
                "\n                - dependencies"
                "\n                - sklearn_equivalent"
                "\n        show_labels: false"
                "\n        group_by_category: false"
            )
            for avg_method in all_avg_methods.values()
        ),
    )

    logger.info(
        "Averaging - Filled parameter `averaging_methods_list`",
    )

    # Write the template to a md file
    output_fp = AVERAGING_SECTION.resolve()
    output_fp.write_text(
        str(template),
        encoding="utf-8",
    )

    logger.info(
        f"Averaging - Wrote filled template to '{output_fp}'",
    )


def experiment_aggregation() -> None:
    # Load in the template
    template_fp = TEMPLATE_DIR / "experiment_aggregation.md"
    if not template_fp.exists():
        raise FileNotFoundError(f"Could not find a template file at '{template_fp}'")

    logger.info(f"Experiment Aggregation - Found template at '{template_fp}'")

    template = Template(
        file_name=template_fp,
    )

    # Complete the template
    # Aliases table ============================================================
    # Creates a table with some important information as an overview
    methods = []
    for alias, method in AGGREGATION_REGISTRY.items():
        methods.append(
            [
                f"'{alias}'",
                f"[{method.__name__}](#{method.__module__}.{method.__name__})",
            ],
        )

    methods = sorted(methods, key=lambda x: x[0])

    table_str = tabulate(
        tabular_data=methods,
        headers=["Alias", "Method"],
        tablefmt="github",
    )

    template.set(key="experiment_aggregators_table", value=table_str)

    logger.info("Experiment Aggregation - Filled 'experiment_aggregators_table'")

    # Write the template to an md file
    output_fp = EXPERIMENT_AGGREGATION_SECTION.resolve()
    output_fp.write_text(
        str(template),
        encoding="utf-8",
    )

    logger.info(
        f"Experiment Aggregation - Wrote filled template to '{output_fp}'",
    )


if __name__ == "__main__":
    logging.basicConfig(
        stream=sys.stdout,
        level=logging.INFO,
        format="[%(funcName)s] %(message)s",
    )

    # References/Metrics/Metrics.md
    metrics()

    # References/Metrics/Averaging.md
    averaging()

    # References/Metrics/index.md
    metrics_and_averaging_overview()

    # References/Experiment Aggregation/index.md
    experiment_aggregation()



import numpy as np
import pandas as pd
import pytest

from prob_conf_mat import Study


@pytest.fixture(scope="module")
def study() -> Study:
    totsch_table_3 = [
        [1, "3467175", 0.99763, 15087, 36, 7544, 18, 18, 7544],
        [2, "3394520", 0.99672, 15073, 50, 7537, 25, 25, 7537],
        [3, "3338942", 0.99596, 15062, 61, 7531, 31, 31, 7531],
        [4, "3339018", 0.99512, 15049, 74, 7525, 37, 37, 7525],
        [5, "3338836", 0.99498, 15047, 76, 7524, 38, 38, 7524],
        [6, "3429037", 0.9938, 15029, 94, 7515, 47, 47, 7515],
        [7, "3346448", 0.99296, 15017, 106, 7509, 53, 53, 7509],
        [8, "3338664", 0.99296, 15017, 106, 7509, 53, 53, 7509],
        [9, "3338358", 0.99282, 15014, 109, 7507, 55, 55, 7507],
        [10, "3339624", 0.9924, 15008, 115, 7504, 58, 58, 7504],
    ]

    totsch_table_3_df = pd.DataFrame.from_records(
        totsch_table_3,
        columns=["Rank", "TeamId", "Score", "TP+TN", "FP+FN", "TP", "FN", "FP", "TN"],
        index="Rank",
    )

    study = Study(
        seed=0,
        num_samples=100000,
        ci_probability=0.95,
    )

    study.add_metric(metric="acc")

    for rank, row in totsch_table_3_df.iterrows():
        study.add_experiment(
            experiment_name=f"{rank}/{row.TeamId}",
            confusion_matrix=[[row.TP, row.FN], [row.FP, row.TN]],
            confusion_prior=1,
            prevalence_prior=1,
        )

    return study


def test_list_wise_ranks(study) -> None:
    """For this trivial example, make sure that the listwise ordering sees consecutive decreases."""

    listwise_experiment_result = study.get_listwise_comparsion_result(
        metric="acc",
    )

    metric_result: pd.DataFrame = study.report_metric_summaries(
        metric="acc",
        table_fmt="pd",
    )  # type: ignore

    metric_result = metric_result.set_index("Group")

    experiment_ranks = [
        int(experiment_name.split("/")[0])
        for experiment_name in listwise_experiment_result.experiment_names
    ]

    for rank_l, rank_r in zip(experiment_ranks[:-1], experiment_ranks[1:]):
        metric_diff = (
            metric_result.loc[str(rank_l)].Median
            - metric_result.loc[str(rank_r)].Median
        )

        assert metric_diff >= 0.0, (experiment_ranks, rank_l, rank_r, metric_diff)  # type: ignore


def test_expected_reward_computation(study) -> None:
    study.report_expected_reward(metric="acc")
    study.report_expected_reward(metric="acc", rewards=1.0)
    study.report_expected_reward(metric="acc", rewards=np.ones((10,)))

    # Test too many rewards
    with pytest.raises(ValueError):
        study.report_expected_reward(metric="acc", rewards=np.ones((11,)))

    # Test rewards shape
    with pytest.raises(ValueError):
        study.report_expected_reward(metric="acc", rewards=np.ones((2, 10, 2)))



from pathlib import Path

from prob_conf_mat import Study
from prob_conf_mat.io import load_csv

if __name__ == "__main__":
    study = Study(
        seed=0,
        num_samples=10000,
        ci_probability=0.95,
    )

    study.add_metric(metric="acc", aggregation="fe_gaussian")
    study.add_metric(metric="f1", aggregation="fe_gaussian")
    study.add_metric(metric="f1@weighted", aggregation="fe_gaussian")
    study.add_metric(metric="f1@macro", aggregation="fe_gaussian")
    study.add_metric(metric="mcc", aggregation="beta")

    # Iterate over all found csv files
    conf_mat_paths = Path(
        "/home/ioverho/prob_conf_mat/documentation/Getting Started/mnist_digits",
    )
    for file_path in sorted(conf_mat_paths.glob("*.csv")):
        # Split the file name to recover the model and fold
        model, fold = file_path.stem.split("_")

        # Load in the confusion matrix using the utility function
        confusion_matrix = load_csv(location=file_path)

        # Add the experiment to the study
        study.add_experiment(
            experiment_name=f"{model}/fold_{fold}",
            confusion_matrix=confusion_matrix,
            prevalence_prior="ones",
            confusion_prior="zeros",
        )

    # Report on the aggregate F1 macro scores
    study.report_aggregated_metric_summaries(metric="f1@macro")

    # Plot the aggregated F1 macro scores for all experiment groups
    study.plot_forest_plot(metric="f1@macro")

    # Compare the aggregated F1 macro scores
    study.plot_pairwise_comparison(
        metric="f1@macro",
        experiment_a="mlp/aggregated",
        experiment_b="svm/aggregated",
        min_sig_diff=0.005,
    )

    # Plot the difference in the aggregated F1 macro scores
    study.plot_pairwise_comparison(
        metric="f1@macro",
        experiment_a="mlp/aggregated",
        experiment_b="svm/aggregated",
        min_sig_diff=0.005,
    )

    # Dump the study config
    study_config = study.to_dict()



from __future__ import annotations
import typing

if typing.TYPE_CHECKING:
    import jaxtyping as jtyping

from timeit import timeit

import numpy as np
from opt_einsum import contract
from tqdm import tqdm

from prob_conf_mat.utils.rng import RNG


def normalize_array_baseline(  # noqa
    arr: jtyping.Float[np.ndarray, "... num_classes"],
) -> jtyping.Float[np.ndarray, "... num_classes"]:
    return arr / np.sum(arr, axis=-1, keepdims=True)


def normalize_array_1(  # noqa
    arr: jtyping.Float[np.ndarray, "... num_classes"],
) -> jtyping.Float[np.ndarray, "... num_classes"]:
    np.divide(arr, np.sum(arr, axis=-1, keepdims=True), out=arr)
    return arr


def normalize_array_2(  # noqa
    arr: jtyping.Float[np.ndarray, "... num_classes"],
) -> jtyping.Float[np.ndarray, "... num_classes"]:
    arr /= np.sum(a=arr, axis=-1, keepdims=True)
    return arr


def normalize_array_3(  # noqa
    arr: jtyping.Float[np.ndarray, "... num_classes"],
) -> jtyping.Float[np.ndarray, "... num_classes"]:
    sum_arr = np.einsum("...i->...", arr)
    arr /= sum_arr[..., np.newaxis]
    return arr


def normalize_array_4(  # noqa
    arr: jtyping.Float[np.ndarray, "... num_classes"],
) -> jtyping.Float[np.ndarray, "... num_classes"]:
    np.divide(arr, np.einsum("...i->...", arr)[..., np.newaxis], out=arr)
    return arr


def normalize_array_5(  # noqa
    arr: jtyping.Float[np.ndarray, "... num_classes"],
) -> jtyping.Float[np.ndarray, "... num_classes"]:
    sum_arr = contract("...i->...", arr)
    arr /= sum_arr[..., np.newaxis]
    return arr


def normalize_array_6(  # noqa
    arr: jtyping.Float[np.ndarray, "... num_classes"],
) -> jtyping.Float[np.ndarray, "... num_classes"]:
    np.divide(arr, contract("...i->...", arr)[..., np.newaxis], out=arr)
    return arr


def normalize_array_7(  # noqa
    arr: jtyping.Float[np.ndarray, "... num_classes"],
) -> jtyping.Float[np.ndarray, "... num_classes"]:
    sum_arr = np.empty((arr.shape[0], arr.shape[1]))
    contract("...i->...", arr, out=sum_arr)
    arr /= sum_arr[..., np.newaxis]
    return arr


def normalize_array_8(  # noqa
    arr: jtyping.Float[np.ndarray, "... num_classes"],
) -> jtyping.Float[np.ndarray, "... num_classes"]:
    sum_arr = contract("...i->...", arr, optimize="greedy")
    arr /= sum_arr[..., np.newaxis]
    return arr


def normalize_array_9(  # noqa
    arr: jtyping.Float[np.ndarray, "... num_classes"],
) -> jtyping.Float[np.ndarray, "... num_classes"]:
    sum_arr = contract("...i->...", arr, optimize="optimal")
    arr /= sum_arr[..., np.newaxis]
    return arr


def normalize_array_10(  # noqa
    arr: jtyping.Float[np.ndarray, "... num_classes"],
) -> jtyping.Float[np.ndarray, "... num_classes"]:
    sum_arr = np.empty((arr.shape[0], arr.shape[1]))
    sum_arr = np.einsum("...i->...", arr)
    arr /= sum_arr[..., np.newaxis]
    return arr


ALL_NORMALIZATION_FUNCS = [
    # normalize_array_1,
    # normalize_array_2,
    normalize_array_3,
    normalize_array_4,
    normalize_array_5,
    # normalize_array_6,
    normalize_array_7,
    # normalize_array_8,
    # normalize_array_9,
    normalize_array_10,
]

if __name__ == "__main__":
    repeats = 10
    iterations = 1000
    num_samples = 10000
    alphas = np.ones((10, 10), dtype=np.float64)

    rng = RNG(seed=0)

    # Broadcast alphas to the desired shape
    alphas = np.broadcast_to(alphas, (num_samples, *alphas.shape))

    # Generate independent gamma-distributed samples
    arr = rng.standard_gamma(alphas)

    all_pcall = []
    for i in tqdm(range(repeats + 1)):
        total_baseline_time = timeit(
            "normalize_array_baseline(arr=arr)",
            number=iterations,
            globals=globals(),
        )

        baseline_normalized_arr = normalize_array_baseline(arr=arr)

        all_times = [total_baseline_time]
        for func in tqdm(ALL_NORMALIZATION_FUNCS):
            total_time = timeit(
                f"{func.__name__}(arr=arr)",
                number=iterations,
                globals=globals(),
            )

            normalized_arr = func(arr=arr)

            assert np.allclose(baseline_normalized_arr, normalized_arr)

            all_times.append(total_time)

        if i != 0:
            all_pcall.append(all_times)

    all_pcall = np.array(all_pcall)
    print(np.median(all_pcall, axis=0))

    for i, func in enumerate([normalize_array_baseline] + ALL_NORMALIZATION_FUNCS):
        print(
            f"{func.__name__:<30} |"
            f" {np.median(all_pcall[:, i]):.2f}"
            f" {np.median(all_pcall[:, i] / iterations):.2e}",
            f" {np.std(all_pcall[:, i]):.2f}",
            f" {np.std(all_pcall[:, i] / iterations):.2e}",
        )
